#!/usr/bin/perl -w
# 
# Generate graphs of the average PPD per WU in the queue
# 

use strict;
use Cwd;
use File::Basename;
use Date::Calc;
use Date::Manip;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";
use GD::Graph::bars;
use GD::Graph::lines;
use RRDs;
use Date::Format qw(time2str);
use POSIX qw(mktime);

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

my $www_data_dir     = '/var/www/fci/';
my $client_data_dir  = 'client-data/';
my $xml_data_dir     = 'xml-data/';
my $queue_graphs_dir = 'queue-graphs/';
my $rrd_data_dir     = 'rrd-data/';

my %cfg = ( self                     => 'fci-generate-queue-graphs.pl',
            version                  => '0.3',
	    
	    verbose                  => 0,
	    quiet                    => 0,
	    
	    'www-data-dir'           => $www_data_dir,
	    'client-data-dir'        => $client_data_dir,
	    'rrd-data-dir'           => $rrd_data_dir,
	    'xml-data-dir'           => $xml_data_dir,
            
            'client-list-file'       => $xml_data_dir.'client-list.xml',
	    'queue-graphs-file'      => $xml_data_dir.'queue-graphs.xml',

	    'action'                 => '',
	    'client'                 => '',
          );

# RRD default configuration
my $rrd_step = 3600;                    # 1 update every 1 hours (3600 seconds)

my %ds_names = ( 
		 'client-queue' => [ 'index_ppd', 'index_speed', 'index_progress', 'index_flops', 'queue_pf', 'queue_ppd' ],
               );
my $ds_heartbeat = $rrd_step * 2;       # 1 update every 1 hours (3600 seconds), and 1 update may be missed.
my $ds_min = 0;                         # Minimum value for every update, cannot be less than zero
my $ds_max = 'U';                       # Maximum value for every update, cannot be more than unlimited

my @rra_cf  = qw(AVERAGE MIN MAX LAST); # Use the average value of the PDPs to form the CDP
my $rra_xff = '0.5';                    # Half the PDPs may be *UNKNOWN*

my $rra_day_steps = 1;                  #    $rra_day_steps  (1) x $rrd_step      (3600) =     3600   [1h]
my $rra_day_rows  = 24;                 #              3600 [1h] x $rra_day_rows    (24) =    86400   [1d]

my $rra_week_steps = 24;                #   $rra_week_steps (24) x $rrd_step      (3600) =    86400   [1d]
my $rra_week_rows  = 7;                 #             86400 [1d] x $rra_week_rows    (7) =   604800   [1w]

my $rra_month_steps = 24;               #  $rra_month_steps (24) x $rrd_step      (3600) =    86400   [1d]
my $rra_month_rows  = 31;               #             86400 [1d] x $rra_month_rows  (31) =  2678400  [31d] ~1month

my $rra_year_steps = 24;                #   $rra_year_steps (24) x $rrd_step      (3600) =    86400   [1d]
my $rra_year_rows  = 366;               #             86400 [1d] x $rra_year_rows  (366) = 31622400 [366d] ~1year

my @periods      = qw(1day 1week 1month 1year);

my %period_start = ( '1day'   => 'end-24hours',    # -1day
                     '1week'  => 'end-168hours',   # -7days
                     '1month' => 'end-744hours',   # -31days
                     '1year'  => 'end-8784hours',  # -366days
                   );

# Get commandline parameters
Getopt::Mixed::init("dir=s
                     client=s
                     verbose
		     quiet
		     help
		     d>dir
		     c>client
		     v>verbose
		     q>quiet
		     h>help
		    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit 0;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{'client'}       = $value if($option eq 'client');
	
	$cfg{verbose}    = 1      if($option eq 'verbose');
	$cfg{quiet}      = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# if quiet was requested together with verbose, disable verbose which contradicts quiet
if($cfg{quiet} && $cfg{verbose}) {
	#print "--quiet overrides --verbose\n";
	$cfg{verbose} = 0;
}

# add directory separator to the end of the www-data-dir path if missing
$cfg{'www-data-dir'} .= '/' if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}


# show what's print()ed without waiting for the newline $|=1;
$|=1;


# check file existance and permissions
# TODO

# read the xml files
my $client_list         = read_xml(%cfg, Name => 'client-list')    || die "Error: Cannot read client-list XML file!\n";
my $queue_graphs        = read_xml(%cfg, Name => 'queue-graphs')   || die "Error: Cannot read queue-graphs XML file!\n";

#print Dumper($config);
#print Dumper($queue_graphs);

if($cfg{client} eq '') {
	# For each clients processed qd-data.xml file, 
	# generate a graph with the average PPD for the WUs in the queue.dat
	
	print "All clients mode.\n\n" if($cfg{verbose});

	foreach my $client (keys %{$client_list->{client_list}->{client}}) {
	        next if($client !~ /^[a-zA-Z0-9\_\-]+$/);
		
		print "Processing client: $client\n" if($cfg{verbose});

		if(defined $client_list->{client_list}->{client}->{$client}->{state} &&
		   $client_list->{client_list}->{client}->{$client}->{state} eq "inactive"
		) {
			print "Inactive client, skipping.\n\n" if($cfg{verbose});
			next;
		}
		else {
			print "Active client, generating graphs.\n" if($cfg{verbose});
		}

		if(has_qd_data($client, $client_list, %cfg)) {
			generate_ppd_graph_for_client($client, %cfg);

			generate_rrd_graphs_for_client($client, %cfg);
		}
		else {
			print "No usable qd-data.xml for client: $client, skipping this client.\n" if($cfg{verbose});
			next;
		}
	
		print "\n" if($cfg{verbose});
	}
}
else {
	print "Single client mode.\n\n" if($cfg{verbose});
	
	if($cfg{client} =~ /^[a-zA-Z0-9\_\-]+$/) {
		print "Processing client: $cfg{client}\n" if($cfg{verbose});
	
		if(has_qd_data($cfg{client}, $client_list, %cfg)) {
			generate_ppd_graph_for_client($cfg{client}, %cfg);

			generate_rrd_graphs_for_client($cfg{client}, %cfg);
		}
		else {
			print "Error: No usable qd-data.xml for client: $cfg{client}\n";
			exit 1;
		}
	}
}

print "Saving updated queue-graphs XML file ...\n" if($cfg{verbose});
my ($success, $reason) = write_xml( %cfg, 
                                    Name => 'queue-graphs',
				    XML  => $queue_graphs
				  );

if(!$success) {
	print "Error: Failed to write queue-graphs XML file!\n";
	exit 1;
}


exit 0; # exit normally

################################[ Subroutines ]################################

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}


##############
#
#     Routine: read_file()
#  Parameters: path to file
#     Returns: content of the file
# 	       undef in case of error
# Description: return the contents of a file
#

sub read_file {
	my $file    = $_[0];
	my $verbose = $_[1] ? $_[1] : 0;
	
	print "Reading file from disk: $file ... " if($verbose);
	if(-r $file) {
		my $data = '';

		open(F, "< $file") || return undef;
		while(<F>) {
			$data .= $_;
		}
		close F;

		print "Done.\n" if($verbose);

		return $data;
	}
	else {
		print "FAILED!\n" if($verbose);
		return undef;
	}
}


##############
#
#     Routine: save_file()
#  Parameters: path to file
# 	       data to save into the file
# 	       optional hash with
# 	         mode    => file premission in octal
# 	         binary  => flag to set binary mode (only needed on Windows)
# 	         verbose => flag to set verbose output
#     Returns: 1 for successfull save or undef for unsuccessfull save
#              the reason for the failure or an empty string
# Description: write a file to disk if the required permissions are set
#

sub save_file {
	my $filename = shift;
	my $content  = shift;
	my %option   = @_;

	my ($success, $reason) = (1, '');

	print "Saving file to disk: $filename ... " if($option{verbose});
	if(!-e $filename || -w $filename) {
		print "Done.\n" if($option{verbose});
		
		open(F, "> $filename") || return (0, "Error: Cannot save file: $filename ($!)\n"); 
		if($option{binary} && $option{binary} == 1) {
			print "Binmode enabled.\n" if($option{verbose});
			binmode(F);
		}
		print F $content;
		close F;

		if($option{mode} && $option{mode} =~ /^0\d{3}$/) {
			print "Set permission: $option{mode}.\n" if($option{verbose});
			chmod oct($option{mode}), $filename;
		}
		else {
			if($option{verbose}) {
				print "Ignoring invalid mode: ";
				print $option{mode};
				print "\n";
			}
		}

		return 1;
	}
	else {
		print "FAILED! (File is not writeable)\n" if($option{verbose});
		return (0, "Error: File does not exists or is unwriteable!\n");
	}
}


##############
#
#     Routine: resize_image()
#  Parameters: source image
#              target image
#              width in pixels
#              height in pixels
#     Returns: 1 on success
#              undef in case of error
# Description: Resize an image with Image::Magick
#

sub resize_image {
        my $source_file  = $_[0];
        my $target_file  = $_[1];
	my $width  	 = $_[2];
	my $height 	 = $_[3];
	my $blur   	 = 1;
	
	if(-r $source_file && (!-e $target_file || -w $target_file)) {
		my $image = Image::Magick->new(magick => 'jpg');

		$image->Read($source_file);

		my ($img_width, $img_height) = $image->Get('width','height');

		if($img_width > $img_height) {
			# Landscape orientated image

			if($img_width > $width) {
				# image is bigger that target size, so resize it
				
				my $new_width  = $width;
				my $new_height = $img_height / ($img_width / $new_width);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		elsif($img_height > $img_width) {
			# portrait orientated image
			
			if($img_height > $height) {
				# image is bigger that target size, so resize it

				my $new_height = $height;
				my $new_width  = $img_width / ($img_height / $new_height);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		else {
			# square image (width eq heigth)
			
			if($img_height > $height) {
				# image is bigger that target size, so resize it

				my $new_height = $height;
				my $new_width  = $img_width / ($img_height / $new_height);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
			elsif($img_width > $width) {
				# image is bigger that target size, so resize it
				
				 my $new_width  = $width;
				 my $new_height = $img_height / ($img_width / $new_width);

				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
			else {
				# image is bigger that target size, so resize it

				my $new_width  = $width;
				my $new_height = $img_height / ($img_width / $new_width);
				
				$image->Resize(
					height => $new_height,
					width  => $new_width,
					blur   => $blur
				);
			}
		}
		
		# write resized file to disk
		$image->Write( filename => $target_file );
		chmod(0664, $target_file);
		
		return 1;
	}
	else {
		return undef;
	}
}

##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
                elsif($args{Name} eq 'jmol-missing-projects') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'project' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
                elsif($args{Name} eq 'queue-graphs') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'client' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
                elsif($args{Name} eq 'queue-graphs') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
	}

	return ($success, $reason);
}

sub get_ppd {
	my $current_index = shift;
	my $qd_data       = shift;
	
	my %info;
	if(exists $qd_data->{queue}->{index}->{"$current_index"}->{project_number} &&
	   $qd_data->{queue}->{index}->{"$current_index"}->{project_number} =~ /^\d+$/
	) {
		$info{project_number} = $qd_data->{queue}->{index}->{"$current_index"}->{project_number};

		#print "Found project number: $info{project_number}\n" if($cfg{verbose});
		
		# get run, clone & generation numbers
		if(exists $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{run} &&
		   exists $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{clone} &&
		   exists $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{generation} &&
		   $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{run} =~ /^\d+$/ &&
		   $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{clone} =~ /^\d+$/ &&
		   $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{generation} =~ /^\d+$/
		) {
			$info{run} = $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{run};
			$info{clone} = $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{clone};
			$info{generation} = $qd_data->{queue}->{index}->{"$current_index"}->{folding}->{generation};
		
			#print "Found Run, Clone & Generation number: $info{run}, $info{clone}, $info{generation}\n" if($cfg{verbose});
		
			# get (credit|points)_per_hour and convert to points_per_day
			if(exists $qd_data->{queue}->{index}->{"$current_index"}->{credit_per_hour} && 
			   $qd_data->{queue}->{index}->{"$current_index"}->{credit_per_hour} =~ /^\d+\.\d+$/
			) {
				$info{ppd} = $qd_data->{queue}->{index}->{"$current_index"}->{credit_per_hour} * 24;
				
				#print "Found PPD: $info{ppd}\n" if($cfg{verbose});
			}
			elsif(exists $qd_data->{queue}->{index}->{"$current_index"}->{points_per_hour} &&
			      $qd_data->{queue}->{index}->{"$current_index"}->{points_per_hour} =~ /^\d+\.\d+$/
			) {
				$info{ppd} = $qd_data->{queue}->{index}->{"$current_index"}->{points_per_hour} * 24;
				
				#print "Found PPD: $info{ppd}\n" if($cfg{verbose});
			}
		}
	}

	return %info;
}

sub generate_ppd_graph_for_client {
	my $client_name = shift;
	my %cfg         = @_;

	my $qd_file = $client_list->{client_list}->{client}->{$client_name}->{files}->{qd_data};

        # qd-data.xml content
	print "Parsing qd-data XML file: $qd_file ... \n" if($cfg{verbose});
        my $qd_data = read_xml( %cfg, 
				Name => 'qd',
				File => $qd_file
			      ) || die "Failed to read project-images XML file!\n";

	#print Dumper($qd_data);
	
	my $current_index = -1;
	if($qd_data->{queue}->{current_index} && $qd_data->{queue}->{current_index} ne '') {
		$current_index = $qd_data->{queue}->{current_index};
	}
	
	# Start with the oldest WU
	if($current_index == 9) {
		$current_index = 0;
	}
	else {
		$current_index += 1;
	}
	
	my %project_ppd;
	my @project = ();
	my @ppd     = ();
	my $i       = 0;
	for(my $index = $current_index ; $index <= 9; $index++) {
		if(exists $qd_data->{queue}->{index}->{"$index"}->{project_number} &&
		   $qd_data->{queue}->{index}->{"$index"}->{project_number} =~ /^\d+$/
		) {
			%project_ppd = get_ppd($index, $qd_data);
			
			my $prcg  = sprintf("%4d", $project_ppd{project_number});
			   $prcg .= " ($project_ppd{run}, $project_ppd{clone}, $project_ppd{generation})";
		
			if($project_ppd{ppd}) {
				if($cfg{verbose}) {
					print "Adding PPD: ";
					printf("%8s", $project_ppd{ppd});
					print "  for: $prcg\n";
				}
			   
				push @project, $prcg;
				push @ppd, sprintf("%.2f", $project_ppd{ppd});
			}
		}
	}
	for(my $index = 0; $index < $current_index; $index++) {
		if(exists $qd_data->{queue}->{index}->{"$index"}->{project_number} &&
		   $qd_data->{queue}->{index}->{"$index"}->{project_number} =~ /^\d+$/
		) {
			%project_ppd = get_ppd($index, $qd_data);
		
			my $prcg  = sprintf("%4d", $project_ppd{project_number});
			   $prcg .= " ($project_ppd{run}, $project_ppd{clone}, $project_ppd{generation})";
			
			if($project_ppd{ppd}) {
				if($cfg{verbose}) {
					print "Adding PPD: ";
					printf("%8s", $project_ppd{ppd});
					print "  for: $prcg\n";
				}
			   
				push @project, $prcg;
				push @ppd, sprintf("%.2f", $project_ppd{ppd});
			}
		}
	}

	my $graph_file = $queue_graphs_dir.$client_name.'.png';
	my $width      = 500;   # GoT Max: 570
	my $height     = 400;   # GoT Max: 430

	my @x = @project;
	my @y = @ppd;
	my @l = @project; # legend

	my $max = 0;
	foreach(@ppd) {
		if($_ > $max) {
			$max = $_;
		}
	}
	$max = int($max + 50);

	if(@x && @y) {
		my $date = get_date();
	
	        my @data  = (\@x, \@y);
	        my $graph = GD::Graph::bars->new($width, $height);
	
	        my $data = GD::Graph::Data->new(\@data);
       		my $values = $data->copy;
		   
		   $graph->set(show_values => $values);
		   $graph->set(dclrs => [ qw(black) ] );
		   
		   $graph->set( x_label           => "Project (Run, Clone, Generation)",
		   		y_label           => "PPD",
				title             => "Average PPD Of Work Units In The Queue",
				x_labels_vertical => 1,
				y_max_value       => $max,
				
				fgclr             => '#000000',	
				bgclr             => '#FFFFFF',	
				labelclr          => '#000000',
				axislabelclr      => '#000000',
				legendclr         => '#000000',
				valuesclr         => '#000000',
				textclr           => '#000000',
			      ) or die $graph->error;
			
		my $gd = $graph->plot(\@data) or die $graph->error;
		
		my ($success, $reason) = save_file( $graph_file, 
		                                    $graph->plot(\@data)->png, 
						    mode => "0664", verbose => $cfg{verbose}
						  );
		if($success) {
			# add the graph file for this client to the queue-graphs list
			$queue_graphs->{queue_graphs}->{client}->{$client_name}->{file} = $graph_file;
			$queue_graphs->{queue_graphs}->{client}->{$client_name}->{last_update} = $date;
			$queue_graphs->{queue_graphs}->{client}->{$client_name}->{state} 
				= $client_list->{client_list}->{client}->{$client_name}->{state};

			print "Successfully generated graph for client: $client_name\n" if($cfg{verbose});
		}
		else {
			print $reason;
			exit 1;
		}
	}
	else {
		print "No data found to generate graph for client: $client_name.\n" if($cfg{verbose});
	}
}

sub generate_rrd_graphs_for_client {
	my $client_name = shift;
	my %cfg         = @_;

	my $qd_file = $client_list->{client_list}->{client}->{$client_name}->{files}->{qd_data};

        # qd-data.xml content
	print "Parsing qd-data XML file: $qd_file ... \n" if($cfg{verbose});
        my $qd_data = read_xml( %cfg, 
				Name => 'qd',
				File => $qd_file
			      ) || die "Failed to read project-images XML file!\n";

	#print Dumper($qd_data);
	
	my $current_index  = $qd_data->{queue}->{current_index};
	if( !($current_index >= 0 && $current_index <= 9) ) {
		print "Invalid current index found: $current_index, skipping RRD update.\n" if($cfg{verbose});
		return;
	}
	
	my $index_ppd = 'U';
	if($qd_data->{queue}->{index}->{$current_index}->{points_per_hour}) {
		$index_ppd  = $qd_data->{queue}->{index}->{$current_index}->{points_per_hour};
	   	$index_ppd *= 24;
	}
	
	my $index_speed = 'U';
	if($qd_data->{queue}->{index}->{$current_index}->{progress_rate}) {
		$index_speed = $qd_data->{queue}->{index}->{$current_index}->{progress_rate};
	}
	
	my $index_progress = 'U';
	if($qd_data->{queue}->{index}->{$current_index}->{progress}) {
		$index_progress = $qd_data->{queue}->{index}->{$current_index}->{progress};
		$index_progress =~ s/%$//;
	}
	
	my $index_flops = 'U'; 
	if($qd_data->{queue}->{index}->{$current_index}->{flops}) {
		$index_flops = $qd_data->{queue}->{index}->{$current_index}->{flops};
	}
	
	my $queue_pf = 'U';
	if($qd_data->{queue}->{performance_fraction}->{fraction}) {
		$queue_pf = $qd_data->{queue}->{performance_fraction}->{fraction};
	}

	my $queue_ppd = 'U';
	if($qd_data->{queue}->{ppd}) {
		$queue_ppd = $qd_data->{queue}->{ppd};
	}

	my $executed = time();
	if($qd_data->{executed_utc} && $qd_data->{executed_utc} =~ /^\w+\s+(\w+)\s+(\d+)\s+(\d{2}):(\d{2}):(\d{2})\s+UTC\s+(\d{4})$/) {
		# Thu Apr 09 17:18:43 UTC 2009
		my $month = $1;
		my $day   = $2;
		my $hour  = $3;
		my $min   = $4;
		my $sec   = $5;
		my $year  = $6;

		$month = Date::Calc::Decode_Month($month);
		$month--;

		$year -= 1900;

		$executed = mktime($sec, $min, $hour, $day, $month, $year);
	}

	my $time = $executed;
	
	my @data = ($client_name, $time, $index_ppd, $index_speed, $index_progress, $index_flops, $queue_pf, $queue_ppd);
        
	update_rrd('client-queue', \@data, %cfg);
}


sub has_qd_data {
	my $client      = shift;
	my $client_list = shift;
	my %cfg         = @_;

	if(!exists $client_list->{client_list}->{client}->{$client}->{files}->{qd_data}) {
		if($cfg{verbose}) {
			print "No qd-data.xml listed in the client-list file.\n";
		}
		return 0;
	}
	elsif($client_list->{client_list}->{client}->{$client}->{files}->{qd_data} =~ /^\s*$/) {
		if($cfg{verbose}) {
			print "Warning: Empty qd-data listed.\n";
		}
		return 0;
	}
	elsif(!-r $client_list->{client_list}->{client}->{$client}->{files}->{qd_data}) {
		if($cfg{verbose}) {
			print "Warning: Cannot read qd-data.xml file: ";
			print $client_list->{client_list}->{client}->{$client}->{files}->{qd_data};
			print "\n";
		}
		return 0;
	}

	return 1;
}

sub time_format {
	my $value = shift;
	my $ret;
	
	if ($value >= 0)
	{
		# convert time in seconds to human readable format
		# e.g. 65 seconds == 1m5s
		
		my $date1 = ParseDate("today");
		my $date2 = DateCalc($date1, "+ $value seconds");
		
		my $delta = DateCalc($date1, $date2);
		#print "Delta: $delta\n";

		#+0:0:0:0:1:55:43
		if($delta =~ /(\d+):(\d+):(\d+):(\d+):(\d+):(\d+):(\d+)$/) {
			#+Y:+M:+W:+D:+H:+MN:+S
			my $years   = $1;
			my $months  = $2;
			my $weeks   = $3;
			my $days    = $4;
			my $hours   = $5;
			my $minutes = $6;
			my $seconds = $7;
			
			my $time = '';
			if($years > 0) {
				$time .= $years."y";
			}
			if($months > 0) {
				$time .= $months."m";
			}
			if($days > 0) {
				$time .= $days."d";
			}
			if($hours > 0) {
				$time .= $hours."h";
			}
			if($minutes > 0) {
				$time .= $minutes."m";
			}
			if($seconds > 0) {
				$time .= $seconds."s";
			}

			if($cfg{verbose}) {
				#print "Debug: ";
				#print "years: $years ";
				#print "months: $months ";
				#print "weeks: $weeks ";
				#print "days: $days ";
				#print "hours: $hours ";
				#print "minutes: $minutes ";
				#print "seconds: $seconds\n";
				#
				#print "time: $time\n";
			}

			$ret = $time;
		}
		else {
			$ret = $value;
		}
	}
	else
	{
		$ret = $value;
	}
	
	
	return $ret;
}

sub create_rrd {
	my $rrd_name  = shift;
	my $rrd_file  = shift;
	my $rrd_start = shift;
	my %cfg       = @_;

	if($rrd_name eq 'client-queue') {
	        my @ds = ();
	        foreach my $name (@{$ds_names{$rrd_name}}) {
	                push @ds, "DS:$name:GAUGE:$ds_heartbeat:$ds_min:$ds_max";
	        }
	
	        my @rra = ();
	
	        foreach my $cf (@rra_cf) {
	                # RRA for one day   (24 hours)
	                push @rra, "RRA:$cf:$rra_xff:$rra_day_steps:$rra_day_rows";
	
	                # RRA for one week   (7 days)
	                push @rra, "RRA:$cf:$rra_xff:$rra_week_steps:$rra_week_rows";
	
	                # RRA for one month (31 days)
	                push @rra, "RRA:$cf:$rra_xff:$rra_month_steps:$rra_month_rows";
	
	                # RRA for one year (366 days)
	                push @rra, "RRA:$cf:$rra_xff:$rra_year_steps:$rra_year_rows";
	        }
	
	        if($cfg{verbose}) {
			print "Creating RRD file: $rrd_file\n";
			print "Start: $rrd_start (". time2str("%Y-%m-%d %H:%M:%S", $rrd_start) .")\n";
	                print "rrdcreate $rrd_file\n";
	                print "\t--start $rrd_start --step $rrd_step\n";
	                foreach(@ds)  { print "\t$_ \n"; }
	                foreach(@rra) { print "\t$_ \n"; }
	        }
	
	        RRDs::create($rrd_file, '--start', $rrd_start,
	                                '--step',  $rrd_step,
	                                @ds, @rra
	        );
	
	        my $error = RRDs::error;
	        die "Error: $error\n" if $error;
	}
	else {
		print "Warning: Unknown RRD name: $rrd_name, not creating RRD file.\n";
	}
}

sub update_rrd {
	my $rrd_name = shift;
	my $rrd_data = shift;
	my %cfg      = @_;

	if($rrd_name eq 'client-queue') {
		my ($client_name, $time, $index_ppd, $index_speed, $index_progress, $index_flops, $queue_pf, $queue_ppd) = @{$rrd_data};

		my $rrd_file = $cfg{'rrd-data-dir'} . $client_name .'-queue.rrd';

		if(!-e $rrd_file) {
			my $rrd_start = $time - 1; 

			create_rrd($rrd_name, $rrd_file, $rrd_start, %cfg);
		}

                my $last_update = RRDs::last($rrd_file);
                
		if($time > $last_update) {
			my $template = '';
	                my $update   = $time.':';
	
			my %value = ( 'index_ppd'      => $index_ppd,
				      'index_speed'    => $index_speed,
				      'index_progress' => $index_progress,
				      'index_flops'    => $index_flops,
				      'queue_pf'       => $queue_pf,
				      'queue_ppd'      => $queue_ppd,
				    );
		        foreach my $name (@{$ds_names{$rrd_name}}) {
	                        $template .= ':' if($template ne '');
	                        $template .= $name;
	
	                        $update .= ':' if($update ne $time.':');
	                        $update .= $value{$name};
	                }
	
	                if($cfg{verbose}) {
				print "Updating RRD file: $rrd_file\n";
				print "Time: $time (". time2str("%Y-%m-%d %H:%M:%S", $time) .")\n";
				print "Last: $last_update (". time2str("%Y-%m-%d %H:%M:%S", $last_update) .")\n";
	                        print "rrdupdate $rrd_file\n";
	                        print "\t--template $template\n";
	                        print "\t$update\n";
	                }
	
	                RRDs::update($rrd_file, '--template', $template,
	                                        $update
	                );
	
	                my $error = RRDs::error;
	                die "Error: $error\n" if $error;
		}
		else {
			if($cfg{verbose}) {
				print "Not updating RRD file: $rrd_file\n";
				print "Time not newer than the last update: ";
				print "$time (". time2str("%Y-%m-%d %H:%M:%S", $time) .") <= ";
				print "$last_update (". time2str("%Y-%m-%d %H:%M:%S", $last_update) .")\n";
			}
		}
		
		# rrdgraph ...
		graph_rrd($rrd_name, $rrd_file, $rrd_data, %cfg);
	}
	else {
		print "Warning: Unknown RRD name: $rrd_name, not updating RRD file.\n";
	}
}

sub graph_rrd {
	my $rrd_name  = shift;
	my $rrd_file  = shift;
	my $rrd_data  = shift;
	my %cfg       = @_;

	if($rrd_name eq 'client-queue') {
		my ($client_name, $time, $index_ppd, $index_speed, $index_progress, $index_flops, $queue_pf, $queue_ppd) = @{$rrd_data};

	        foreach my $name (@{$ds_names{$rrd_name}}) {
			foreach my $period (@periods) {
				(my $graph_file = $rrd_file) =~ s/\.rrd$/-${name}-${period}.png/;
	
	                        # 300x100 becomes 397x168
	                        # 550x175 becomes 647x254
	                        my $graph_width  = 550; # GoT Max: 570
	                        my $graph_height = 175; # GoT Max: 430
	
	                        my @def     = ();
	                        my @cdef    = ();
	                        my @vdef    = ();
	                        my @line    = ();
	                        my @gprint  = ();
	                        my @comment = ();
	
				my %human_name = ( 'index_ppd'      => 'Current PPD',
						   'index_speed'    => 'Speed',
						   'index_progress' => 'Progress (%)',
						   'index_flops'    => 'Flops',
						   'queue_pf'       => 'Performance Fraction',
						   'queue_ppd'      => 'Average PPD',
						 );
	
				my %color = ( 'average' => '#0000FF',
					      'min'     => '#00FF00',
					      'max'     => '#FF0000',
					      'last'    => '#000000',
					    );
				
		               	push @def,  "DEF:${name}=${rrd_file}:${name}:LAST";
	                       	push @line, "LINE1:${name}#0000FF:$human_name{$name}";
				if($name eq 'index_ppd' ||
				   $name eq 'queue_ppd'
				) {
		                       	push @gprint, "GPRINT:${name}:MIN:Min\\: %.2lf";
		                       	push @gprint, "GPRINT:${name}:MAX:Max\\: %.2lf";
		                       	push @gprint, "GPRINT:${name}:AVERAGE:Avg\\: %.2lf";
		                       	push @gprint, "GPRINT:${name}:LAST:Cur\\: %.2lf";
				}
				elsif($name eq 'index_speed' ||
				      $name eq 'index_progress'
				) {
		                       	push @gprint, "GPRINT:${name}:MIN:Min\\: %.3lf";
		                       	push @gprint, "GPRINT:${name}:MAX:Max\\: %.3lf";
		                       	push @gprint, "GPRINT:${name}:AVERAGE:Avg\\: %.3lf";
		                       	push @gprint, "GPRINT:${name}:LAST:Cur\\: %.3lf";
				}
				elsif($name eq 'index_flops') {
		                       	push @gprint, "GPRINT:${name}:MIN:Min\\: %.0lf";
		                       	push @gprint, "GPRINT:${name}:MAX:Max\\: %.0lf";
		                       	push @gprint, "GPRINT:${name}:AVERAGE:Avg\\: %.0lf";
		                       	push @gprint, "GPRINT:${name}:LAST:Cur\\: %.0lf";
				}
				elsif($name eq 'queue_pf') {
		                       	push @gprint, "GPRINT:${name}:MIN:Min\\: %.6lf";
		                       	push @gprint, "GPRINT:${name}:MAX:Max\\: %.6lf";
		                       	push @gprint, "GPRINT:${name}:AVERAGE:Avg\\: %.6lf";
		                       	push @gprint, "GPRINT:${name}:LAST:Cur\\: %.6lf";
				}
	
	                        my $last_update = RRDs::last($rrd_file);
	
	                        my $time_start = $period_start{$period};
	                        my $time_end   = $last_update;
	
	                        my $c = 'Last Update: '. time2str("%Y-%m-%d %H:%M:%S", $last_update);
	                           $c =~ s/:/\\:/g;
	                        push @comment, "COMMENT:$c";
	
	                        if($cfg{verbose}) {
					print "Graphing RRD file: $rrd_file\n";
					print "Start: $time_start\n";
					print "End  : $time_end (". time2str("%Y-%m-%d %H:%M:%S", $time_end) .")\n";
	                                print "rrdgraph $graph_file\n";
	                                print "\t--start $time_start\n";
	                                print "\t--end $time_end\n";
	                                print "\t--title \"$client_name - $human_name{$name}\"\n";
	                                print "\t--vertical-label \"$human_name{$name}\"\n";
	                                print "\t--width $graph_width\n";
	                                print "\t--height $graph_height\n";
	                                #print "\t--full-size-mode\n"; # requires rrdtool v1.3
	                                print "\t--imgformat PNG\n";
					print "\t--alt-autoscale\n";
	                                foreach(@def)     { print "\t$_\n"; }
	                                foreach(@cdef)    { print "\t$_\n"; }
	                                foreach(@vdef)    { print "\t$_\n"; }
	                                foreach(@line)    { print "\t$_\n"; }
	                                foreach(@gprint)  { print "\t$_\n"; }
	                                foreach(@comment) { print "\t$_\n"; }
	                        }
	
	                        my ($averages,$xsize,$ysize) = RRDs::graph( $graph_file,
	                                                                    '--start', $time_start,
	                                                                    '--end', $time_end,
	                                                                    '--title', "$client_name - $human_name{$name}",
	                                                                    '--vertical-label', "$human_name{$name}",
	                                                                    '--width', $graph_width,
	                                                                    '--height', $graph_height,
	                                                                    #'--full-size-mode', # requires rrdtool v1.3
	                                                                    '--imgformat', 'PNG',
									    '--alt-autoscale',
	                                                                    @def, @cdef, @vdef, @line, @gprint, @comment
	                                                                  );
	
	                        my $error = RRDs::error;
	                        die "Error: $error\n" if $error;
	
	                        if($cfg{verbose}) {
	                                print "Image Size: ${xsize}x${ysize}\n";
	                                print "Averages: ", (join ", ", @$averages);
	                                print "\n";
	                        }
			}
		}
	}
	else {
		print "Warning: Unknown RRD name: $rrd_name, not updating RRD file.\n";
	}
}

sub print_usage {
	my %cfg = @_;
	
	print "$cfg{self} v$cfg{version}\n\n";
	print "Usage: $cfg{self} [OPTIONS]\n\n";
	print "Options:\n";
	print "-d, --dir <DIR>         Path of the FCI server directory\n";
	print "-c, --client <NAME>     Client to generate graphs for\n";
	
	print "-v, --verbose           Enable verbose output\n";
	print "-q, --quiet             Don't print any output at all, except for errors\n";
	print "-h, --help              Show this usage information\n";
}
