#!/usr/bin/perl -w
#
#        Name: fci-update-xml-files.pl
#      Author: Bas Couwenberg <sebastic@xs4all.nl>
#     License: GPL
# Description: Parse the data uploaded by the clients, and the data downloaded 
#              from Stanford, and compile that into serveral XML files.
#              These XML files are used by the web frontend to show the 
#              progress and other information on the clients.
#

use strict;
use Cwd;
use File::Basename;
use File::Copy;
use Date::Calc;
use Date::Manip;
use XML::Simple;
use Data::Dumper;
use Getopt::Mixed "nextOption";
use Benchmark ':hireswallclock';

# set the Date::Calc language
Date::Calc::Language(Date::Calc::Decode_Language('English'));

# files:
#  client.xml        per client xml file
#  client-list.xml   list of clients and their per client file
#
#  username.xml      per user xml files
#  username-list.xml list of usernames and their per user file
#
#  team.xml          per team xml file
#  team-list.xml     list of teams and their per team file
#
#  project.xml       per project xml file
#  project-list.xml  list of projects and their per project file

my %cfg = ();
   $cfg{'self'}                        = 'fci-update-xml-files.pl';
   $cfg{'version'}                     = '0.4';

   $cfg{'www-data-dir'}                = '/var/www/fci/';
   $cfg{'stanford-data-dir'}           = 'stanford-data/';
   $cfg{'client-data-dir'}             = 'client-data/';
   $cfg{'eoc-data-dir'}                = 'eoc-data/';
   $cfg{'xml-data-dir'}                = 'xml-data/';
   
   $cfg{'client-list-file'}            = $cfg{'xml-data-dir'}.'client-list.xml';
   $cfg{'team-list-file'}              = $cfg{'xml-data-dir'}.'team-list.xml';
   $cfg{'project-list-file'}           = $cfg{'xml-data-dir'}.'project-list.xml';
   $cfg{'username-list-file'}          = $cfg{'xml-data-dir'}.'username-list.xml';
   $cfg{'xml-project-summary-file'}    = $cfg{'xml-data-dir'}.'project-summary.xml';
   $cfg{'jmol-projects-file'}          = $cfg{'xml-data-dir'}.'jmol-projects.xml';
   $cfg{'known-projects-file'}         = $cfg{'xml-data-dir'}.'known-projects.xml';
   $cfg{'client-projects-file'}        = $cfg{'xml-data-dir'}.'client-projects.xml';
   $cfg{'fci-summary-file'}            = $cfg{'xml-data-dir'}.'fci-summary.xml';
   $cfg{'eoc-stats-file'}              = $cfg{'xml-data-dir'}.'eoc-stats.xml';
   $cfg{'fahstats-file'}               = $cfg{'xml-data-dir'}.'fahstats.xml';

   $cfg{'user-summary-file'}           = $cfg{'stanford-data-dir'}.'daily_user_summary.txt';
   $cfg{'team-summary-file'}           = $cfg{'stanford-data-dir'}.'daily_team_summary.txt';
   $cfg{'project-summary-file'}        = $cfg{'stanford-data-dir'}.'psummary.html';

   $cfg{'write-xml'}                   = 0;

   $cfg{verbose}                       = 0;
   $cfg{debug}                         = 0;
   $cfg{quiet}                         = 0;

# Get commandline parameters
Getopt::Mixed::init("dir=s
                     verbose
		     debug
		     quiet
		     help
		     d>dir
		     v>verbose
		     b>debug
		     q>quiet
		     h>help
                    ");

while (my ($option, $value) = nextOption()) {
	if($option eq 'help') {
		print_usage(%cfg);
		exit;
	}
	
	$cfg{'www-data-dir'} = $value if($option eq 'dir');
	$cfg{verbose}        = 1      if($option eq 'verbose');
	$cfg{debug}          = 1      if($option eq 'debug');
	$cfg{quiet}          = 1      if($option eq 'quiet');
}
Getopt::Mixed::cleanup();

# add directory separator to the end of the www-root path if missing
if($^O eq 'MSWin32') {
        # Windows
        $cfg{'www-data-dir'} .= '\\' if(substr($cfg{'www-data-dir'}, -1, 1) ne '\\');
}
else {
        # One of the UNIX flavors is assumed
        $cfg{'www-data-dir'} .= '/'  if(substr($cfg{'www-data-dir'}, -1, 1) ne '/');
}

# show what's print()ed without waiting for the newline $|=1;
$|=1;

my $dir = Cwd::cwd;
if(-d $cfg{'www-data-dir'} && -x $cfg{'www-data-dir'}) {
	print "Changing working directory to: $cfg{'www-data-dir'}\n" if($cfg{verbose});
	Cwd::chdir $cfg{'www-data-dir'};
	print "\n" if($cfg{verbose});
}
else {
	print "Error: Cannot change to directory: $cfg{'www-data-dir'}\n";
	exit 1;
}
					

# read the XML files  
print "Reading XML files:\n" if($cfg{verbose});
my $xml_read_start = new Benchmark;

print " $cfg{'xml-project-summary-file'}\n" if($cfg{verbose});
my $time_start = new Benchmark;

my $projects    = XMLin( $cfg{'xml-project-summary-file'},
		         KeyAttr    => 'number',
		         KeepRoot   => 1,
		         ForceArray => [ 'project' ]
		       ) || die $@;

my $time_stop = new Benchmark;
my $time_diff = timediff($time_stop, $time_start);

print "  read in: ". timestr($time_diff) ."\n" if($cfg{verbose});


print " $cfg{'jmol-projects-file'}\n" if($cfg{verbose});
$time_start = new Benchmark;

my $jmol_projects = XMLin( $cfg{'jmol-projects-file'},
			   KeyAttr    => { fah_proj => 'number' },
			   ForceArray => [ 'fah_proj' ],
			   KeepRoot   => 1
			 ) || die $@;

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);

print "  read in: ". timestr($time_diff) ."\n" if($cfg{verbose});


print " $cfg{'known-projects-file'}\n" if($cfg{verbose});
$time_start = new Benchmark;

my $known_projects = XMLin( $cfg{'known-projects-file'},
			    KeyAttr    => { project => 'number', image => 'type' },
			    ForceArray => [ 'project', 'image' ],
			    KeepRoot   => 1
		          ) || die $@;

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);

print "  read in: ". timestr($time_diff) ."\n" if($cfg{verbose});


print " $cfg{'client-list-file'}\n" if($cfg{verbose});
$time_start = new Benchmark;

my $client_list = XMLin( $cfg{'client-list-file'},
                         KeyAttr    => 'name',
			 KeepRoot   => 1,
			 ForceArray => [ 'client' ]
		       ) || die $@;

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);

print "  read in: ". timestr($time_diff) ."\n" if($cfg{verbose});


print " $cfg{'username-list-file'}\n" if($cfg{verbose});
$time_start = new Benchmark;

my $username_list = XMLin( $cfg{'username-list-file'},
                           KeyAttr    => 'userid',
			   KeepRoot   => 1,
			   ForceArray => [ 'username' ]
			 ) || die $@;

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);

print "  read in: ". timestr($time_diff) ."\n" if($cfg{verbose});


print " $cfg{'team-list-file'}\n" if($cfg{verbose});
$time_start = new Benchmark;

my $team_list   = XMLin( $cfg{'team-list-file'},
                         KeyAttr    => 'number',
		         KeepRoot   => 1,
		         ForceArray => [ 'team', 'username' ]
                       ) || die $@;

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);

print "  read in: ". timestr($time_diff) ."\n" if($cfg{verbose});

my $xml_read_stop = new Benchmark;
my $xml_read_diff = timediff($xml_read_stop, $xml_read_start);
print "XML files read in: ". timestr($xml_read_diff) ."\n\n" if($cfg{verbose});

my %client_type = ( '0' => 'No Preference',
		    '1' => 'Regular',
		    '2' => 'No Deadline',
		    '3' => 'Advmethods',
		    '4' => 'Beta',
		    '5' => 'Internal',
		    '6' => 'BigAdv Beta',
		    '7' => 'BigAdv',
		  );

my $fci_summary     = {};

my $team            = {};
my $username        = {};
my $client_projects = {};

# get a list of clients
# for every client:
# - process its data
# - saved processed data in per client xml file, client.xml
# - save subset of client data to client-list xml file

my $start_time = get_date();
$fci_summary->{fci_summary}->{last_update}->{start_time} = $start_time;

my $client_processing_start = new Benchmark;

opendir(DIR, $cfg{'client-data-dir'}) || die "Cannot open directory: $cfg{'client-data-dir'}\n";
foreach my $client_name (readdir DIR) {
	next if($client_name !~ /^[a-zA-Z0-9\_\-]+$/);

	my $client_processing_start = new Benchmark;

	print "Processing client: $client_name\n" if($cfg{verbose});

	my $client_file = $cfg{'client-data-dir'}.$client_name.'/client.xml.new';
	if(!-r $client_file) {
		print "Error: Cannot read new client XML file: $client_file\n" if($cfg{verbose});
		next;
	}

	# client.xml.new content
	my $client = XMLin( $client_file,
	                    ForceArray => [ 'server', 'stats' ]
	                  ) || die $@;
	
	print "\nChecking last update ...\n" if($cfg{verbose});

	my $client_file_prev = $cfg{'client-data-dir'}.$client_name.'/client.xml';
	if(-r $client_file_prev) {
		my $client_prev = XMLin( $client_file_prev,
					 ForceArray => [ 'server', 'stats' ]
				       );

		my $now      = $start_time;
		my $hour_ago = DateCalc($now, "- 1 hour");
		
		if($cfg{verbose}) {
			print "Previous: $client_prev->{last_update}\n" if($client_prev->{last_update});
			print " Current: $client->{last_update}\n";
			print "     Now: $now\n";
		}

		my $date1 = ParseDate($client->{last_update});
		my $date2 = ParseDate($hour_ago);
		my $flag  = Date_Cmp($date1, $date2);
                        
		my $date3 = ParseDate($now);
		my $delta = DateCalc($date3, $date1, {}, 1);
		   $delta = delta_format($delta, 'long');

		# flag <  0: date1 is earlier than date2
                # flag == 0: date1 and date2 are identical
                # flag >  0: date1 is later than date2
		if($flag < 0) {
			# the client upload is more than 1 hour old
			# consider the client inactive, it has missed it's hourly upload

			print "The last upload is more than 1 hour old.\n" if($cfg{verbose});
			$client->{state} = "inactive";

			my $message = '';
			if($delta !~ /^(\+|\-)/) {
				$message = "Last upload: $delta";
			}

			$client->{state_message} = $message;
		}
		elsif($flag == 0) {
			# the client upload is exacly 1 hour old
			# consider the client active for now, he'll likely be inactive with the next run

			print "The last upload is exactly 1 hour old.\n" if($cfg{verbose});
			$client->{state} = "active";
			
			my $message = '';
			if($delta !~ /^(\+|\-)/) {
				$message = "Last upload: $delta";
			}

			$client->{state_message} = $message;
		}
		else {
			# the client upload less than 1 hour old
			# consider the client active, it has done it's hourly upload

			print "The last upload is less than 1 hour old.\n" if($cfg{verbose});
			$client->{state} = "active";
			
			my $message = '';
			if($delta !~ /^(\+|\-)/) {
				$message = "Last upload: $delta";
			}

			$client->{state_message} = $message;
		}

		if(!$client_prev->{last_update}) {
			# The client has not been processed before (new client)
			
			print "This client has not been processed before.\n" if($cfg{verbose});
			$client->{state} = "new";

			my $message = '';
			if($delta !~ /^(\+|\-)/) {
				$message = "First upload: $delta";
			}

			$client->{state_message} = $message;
		}
	}
	if($cfg{verbose}) {
		print "Client is considered: $client->{state}";
		if(exists $client->{state_message} && $client->{state_message} ne '') {
			print " ($client->{state_message})";
		}
		print "\n";
	}
	
	# Replace any pre existing client info from the client-list with the uploaded info,
	# but preserve <first_seen> for clients who have it defined.
	my $first_seen = '';
	if(exists $client_list->{client_list}->{client}->{"$client_name"}) {
		# <first_seen> was introduced in fci-update-xml-files.pl v0.3, if it has not
		# been recorded for clients who were processed with an earlier version, set
		# <first_seen> the clients last update.
		if(defined $client_list->{client_list}->{client}->{"$client_name"}->{first_seen}) {
			$first_seen = $client_list->{client_list}->{client}->{"$client_name"}->{first_seen};
		}
		delete $client_list->{client_list}->{client}->{"$client_name"};
	}
	if($first_seen eq '') {
		$first_seen = $client->{last_update};
	}
	print "Client was first seen on: $first_seen\n\n" if($cfg{verbose});
	$client->{first_seen} = $first_seen;
	
	$client_list->{client_list}->{client}->{"$client_name"}->{files}         = $client->{files};
	$client_list->{client_list}->{client}->{"$client_name"}->{last_update}   = $client->{last_update};
	$client_list->{client_list}->{client}->{"$client_name"}->{state}         = $client->{state};
	$client_list->{client_list}->{client}->{"$client_name"}->{state_message} = $client->{state_message};
	$client_list->{client_list}->{client}->{"$client_name"}->{first_seen}    = $client->{first_seen};

	if($client->{hardware}) {
		$client_list->{client_list}->{client}->{"$client_name"}->{hardware} = $client->{hardware};
	}
	if($client->{operating_system}) {
		$client_list->{client_list}->{client}->{"$client_name"}->{operating_system} = $client->{operating_system};
	}
	if($client->{software}) {
		$client_list->{client_list}->{client}->{"$client_name"}->{software} = $client->{software};
	}

	#print Dumper($client_list); 
	#exit;

	my %username = (); # username.xml content, and username element for username-list.xml

	# add client.xml to files list
	$client->{files}->{client_xml} = 'client-data/'.$client_name.'/client.xml';
	$client_list->{client_list}->{client}->{"$client_name"}->{files}->{client_xml} = $client->{files}->{client_xml};
	
	if(!-w $client->{files}->{client_xml}) {
		print "Error: Cannot write client XML file! ($client->{files}->{client_xml})\n";
		exit 1;
	}

	if($cfg{verbose}) {
		my $longest = 0;
		foreach(keys  %{$client->{files}}) {
			my $length = length($client->{files}->{$_});
	
			$longest = $length if($length > $longest);
		}
				
		print "Files for this client:\n";
		foreach(keys %{$client->{files}}) {
			print " $client->{files}->{$_}";
			print " " x ($longest - length($client->{files}->{$_}));
			print " ($_)\n";
		}
		print "\n";
	}

	# parse: qd-data.xml.new
	my %qddata = ();
	if($client->{files}->{qd_data_new}) {
		my $file = $client->{files}->{qd_data_new};
		
		print "File to parse: $file\n" if($cfg{verbose});
		if(!-r $file) {
			print "Error: Cannot read qd-data.xml file: $file\n";
			exit 1;
		}
		%qddata = parse_qddata($file);
		
		if(!$qddata{error}) {
			# use the data from qd to update the client and client-list XML files
		
			# add all current work unit info to the client specific xml file
			$client->{current_work_unit} = $qddata{queue}->{index}->{$qddata{queue}->{current_index}};
			$client->{current_index}     = $qddata{queue}->{current_index};

			# add average stats from qd to current work unit info
			my @stats = qw /download_rate upload_rate performance_fraction pph ppd ppw ppy/;
			foreach(@stats) {
				if($qddata{queue}->{$_}) {
					$client->{stats}->{$_} = $qddata{queue}->{$_};
				}
			}
		
			# only in qd-data.xml:
			# qd release date
			if($qddata{qd}->{date}) {
				# qd info should only go in the client specific file
				$client->{software}->{qd}->{date} = $qddata{qd}->{date};
			}
			# qd functional revision 
			if($qddata{qd}->{functional_revision}) {
				$client->{software}->{qd}->{functional_revision} = $qddata{qd}->{functional_revision};
			}
			# qdinfo.dat release date and maintainer
			if($qddata{qdinfo}) {
				$client->{software}->{qdinfo} = $qddata{qdinfo};
			}
			# current work unit credit 
			# Note: Stanfords psummary and teamstats files call it credit,
			#       whereas qd calls it points.
			#       In qd 046 and before the XML output of qd uses the tags credit and credit_per_hour
			#       In qd 047 and newer  the XML output of qd uses the tags points and points_per_hour
			# FCI uses the name credit in stead of points internally
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{credit}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{credit} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{credit};
				
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{credit_per_hour}) {
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{credit_per_hour}
						= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{credit_per_hour};
				}
			}	
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{points}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{credit} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{points};

				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{points_per_hour}) {
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{credit_per_hour}
						= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{points_per_hour};
				}

				# $client->{current_work_unit}->{points} should be renamed
				$client->{current_work_unit}->{credit} = $client->{current_work_unit}->{points};
				delete $client->{current_work_unit}->{points};
	
				# $client->{current_work_unit}->{points_per_hour} should be renamed
				if($client->{current_work_unit}->{points_per_hour}) {
					$client->{current_work_unit}->{credit_per_hour} = $client->{current_work_unit}->{points_per_hour};
					delete $client->{current_work_unit}->{points_per_hour};
				}
			}
			# expected time of completion
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{expect_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{expected}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{expect_timestamp};
			}
			# preferred completion deadline
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{preferred_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{preferred}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{preferred_timestamp};
			}
			# completion deadline
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{due_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{deadline}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{due_timestamp};
			}
			# time issued
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{issue_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{issued}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{issue_timestamp};
			}
			# time begun
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{begin_timestamp}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{begun}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{begin_timestamp};
			}
			# folding info (run, clone, gen, ...)
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{folding}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding};
				
				my ($run, $clone, $generation) = (0, 0, 0);
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding}->{run}) {
					$run = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding}->{run};
				}
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding}->{clone}) {
					$clone = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding}->{clone};
				}
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding}->{generation}) {
					$generation = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{folding}->{generation};
				}

				my $trajectory = 'r'. $run .'c'. $clone .'g'. $generation;

				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{trajectory_id} = $trajectory;
			}

			# Detect client operating system if FCI client did not send this data (e.g. pre 1.0 fci client)
			if(!$client_list->{client_list}->{client}->{$client_name}->{operating_system} &&
			   $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{os}->{name}
			) {
				$client_list->{client_list}->{client}->{$client_name}->{operating_system}->{name}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{os}->{name};
				
				$client_list->{client_list}->{client}->{$client_name}->{operating_system}->{version}
					= 'unknown';
				
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{cpu}->{name}) {
					$client_list->{client_list}->{client}->{$client_name}->{operating_system}->{hardware}
						= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{cpu}->{name};
				}
			}

			# check the progress of the current WU
			#  is it expected before the preferred deadline?
			#  is it expected before the final deadline?
			#
			# States:
			#  ok		expected date before all known deadlines (due, and preferred if known)
			#  warning	expected date after preferred deadline
			#  critical     expected date after final deadline (due)
			#  unknown      expected date is not available, or no deadlines available
			
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{expect_timestamp}) {
				my $expect      = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{expect_timestamp};
				   $expect      = iso_date($expect);
				my $expect_date = ParseDate($expect);

				my $preferred      = "";
				my $preferred_date = "";
				my $due            = "";
				my $due_date       = "";
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{preferred_timestamp}) {
					$preferred      = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{preferred_timestamp};
					$preferred      = iso_date($preferred);
				}
				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{due_timestamp}) {
					$due = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{due_timestamp};
					$due = iso_date($due);
				}

				if($preferred ne "" && $due ne "") {
					my $preferred_date  = ParseDate($preferred);
					my $due_date        = ParseDate($due);					

					my $preferred_flag  = Date_Cmp($expect_date, $preferred_date);
					my $due_flag        = Date_Cmp($expect_date, $due_date);

					my $preferred_delta = delta_format(DateCalc($preferred_date, $expect_date, {}, 1), 'long');
					my $due_delta       = delta_format(DateCalc($due_date, $expect_date, {}, 1), 'long');
					
					if($preferred_delta =~ s/ ago$//) {}
					if($due_delta       =~ s/ ago$//) {}

					# flag <  0: date1 is earlier than date2
					# flag == 0: date1 and date2 are identical
					# flag >  0: date1 is later than date2
					
					my $state   = '';
					my $message = '';

					if($preferred_flag < 0 && $due_flag < 0) {
						$state   = 'ok';
						$message = "Expected $preferred_delta before preferred &amp; $due_delta before final deadline";
					}
					elsif($preferred_flag >= 0 && $due_flag < 0) {
						$state   = 'warning';
						$message = "Expected $preferred_delta after preferred &amp; $due_delta before final deadline";
					}
					elsif($preferred_flag >= 0 && $due_flag >= 0) {
						$state   = 'critical';
						$message = "Expected $preferred_delta after preferred &amp; $due_delta after final deadline";
					}
					else {
						$state   = 'unknown';
						$message = "Expected neither before or after preferred &amp neither before or after final deadline";
					}

					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
						= $state;
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state_message}
						= $message;
						
					$client->{current_work_unit}->{progress_state}
						= $state;
					$client->{current_work_unit}->{progress_state_message}
						= $message;
				}
				elsif($due ne "") {
					my $due_date  = ParseDate($due);
					my $due_flag  = Date_Cmp($expect_date, $due_date);
					my $due_delta = delta_format(DateCalc($due_date, $expect_date, {}, 1), 'long');
					if($due_delta =~ s/ ago$//) {}

					my $state   = '';
					my $message = '';

					if($due_flag < 0) {
						$state   = 'ok';
						$message = "Expected $due_delta before final deadline";
					}
					elsif($due_flag >= 0) {
						$state   = 'critical';
						$message = "Expected $due_delta after final deadline";
					}
					else {
						$state   = 'unknown';
						$message = "Expected neither before or after final deadline";
					}

					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
						= $state;
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state_message}
						= $message;
						
					$client->{current_work_unit}->{progress_state}
						= $state;
					$client->{current_work_unit}->{progress_state_message}
						= $message;
				}
				else {
					my $state   = 'unknown';
					my $message = "No preferred nor final deadline known";

					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
						= $state;
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state_message}
						= $message;
						
					$client->{current_work_unit}->{progress_state}
						= $state;
					$client->{current_work_unit}->{progress_state_message}
						= $message;
				}
			}
			else {
				my $state   = 'unknown';
				my $message = "No expect date";

				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state}
					= $state;
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress_state_message}
					= $message;
						
				$client->{current_work_unit}->{progress_state}
					= $state;
				$client->{current_work_unit}->{progress_state_message}
					= $message;
			}
						
			# also in unitinfo.txt:
			# progress
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{progress}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{progress} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{progress};
			}
			# project number
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{project_number}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{project_number}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{project_number};
			}
			# project name
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{project_name}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{project_name} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{project_name};
			}
		
			# also in client.cfg
			# username
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{username}) {
				$client->{config}->{username}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{username};
				
				$username{name} 
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{username};
			}
			# team
			if(exists $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{team} &&
			   $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{team} >= 0
			) {
				my $team_number = $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{team};

				$client->{config}->{team}
					= $team_number;
				
				$team_list->{team_list}->{team}->{$team_number}->{number}
					= $team_number;
				
				$username{team} 
					= $team_number;
			}
			# machineid
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{machine_id}) {
				$client->{config}->{machine_id}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{machine_id};
			}
			# id
			if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{id}) {
				$client->{config}->{id}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{id};
			}
			# type
			if(exists $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{client_type}->{type} &&
			   $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{client_type}->{type} =~ /^\d+$/
			) {
				$client->{config}->{type}
					= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{client_type}->{type};

				if($qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{client_type}->{name}) {
					$client->{config}->{type_name}
						= $qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{client_type}->{name};
				}
				elsif($client_type{$qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{client_type}->{type}}) {
					$client->{config}->{type_name}
						= $client_type{$qddata{queue}->{index}->{$qddata{queue}->{current_index}}->{client_type}->{type}};
				}
			}
		}
		else {
			print "Warning: Error message in the qd-data for client: $client_name!\n";
			print "Error message: $qddata{error}\n\n";

			$client_list->{client_list}->{client}->{$client_name}->{client_error} = "Queue dump error: $qddata{error}";

			$client->{client_error} = "Queue dump error: $qddata{error}";
		}
	
		# add qd-data.xml to the file list of this client
		$client->{files}->{qd_data} = 'client-data/'.$client_name.'/qd-data.xml';
		$client_list->{client_list}->{client}->{"$client_name"}->{files}->{qd_data} = $client->{files}->{qd_data};
		
		# copy qd-data.xml.new to qd-data.xml 
		print "Copying ".basename($file)." to: $client->{files}->{qd_data}\n" if($cfg{verbose});
		copy($file, $client->{files}->{qd_data}) || die "Error: Cannot copy $file -> $client->{files}->{qd_data} ($!)\n";
		chmod(0664, $client->{files}->{qd_data});

		print "\n" if($cfg{verbose});
	}
	else {
		print "Error: No qd-data.xml file for this client!\n";
		print "\nProgram aborted!\n";
		exit 1;
	}

	# parse if available: unitinfo.txt
	my %unitinfo = ();
	if($client->{files}->{unitinfo_new}) {
		my $file = $client->{files}->{unitinfo_new};

		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			# Don't use the unitinfo.txt for data gathering as it is unreliable
			#%unitinfo = parse_unitinfo($file);
                
			# add unitinfo.txt to the file list of this client
			$client->{files}->{unitinfo} = 'client-data/'.$client_name.'/unitinfo.txt';
			$client_list->{client_list}->{client}->{"$client_name"}->{files}->{unitinfo} = $client->{files}->{unitinfo};
		
			# copy unitinfo.txt.new to unitinfo.txt 
			print "Copying ".basename($file)." to: $client->{files}->{unitinfo}\n" if($cfg{verbose});
			copy($file, $client->{files}->{unitinfo}) || die "Error: Cannot copy $file -> $client->{files}->{unitinfo} ($!)\n";
			chmod(0664, $client->{files}->{unitinfo});

			print "\n" if($cfg{verbose});
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{unitinfo_new};
				delete $client->{files}->{unitinfo_new};
			}
			
			print "Error: Cannot read unitinfo.txt file: $file\n";
			exit 1;
		}
	}
	
	# parse if available: client.cfg
	my %clientcfg = ();
	if($client->{files}->{client_cfg_new}) {
		my $file = $client->{files}->{client_cfg_new};
		
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			%clientcfg = parse_clientcfg($file);

			# add client.cfg to the file list of this client
        	        $client->{files}->{client_cfg} = 'client-data/'.$client_name.'/client.cfg';
        	        $client_list->{client_list}->{client}->{"$client_name"}->{files}->{client_cfg} = $client->{files}->{client_cfg};
	
	                # copy client.cfg.new to client.cfg
		        copy($file, $client->{files}->{client_cfg}) || die "Error: Cannot copy $file -> $client->{files}->{client_cfg} ($!)\n";
        	        chmod(0664, $client->{files}->{client_cfg});

			# The username, team & machineID configured in the client.cfg at the time of
			# WU upload are used to credit the WU, not those recorded in the queue.dat
			# at the time the WU was entered in the queue.

			# username
			if($clientcfg{by_key}{username}) {
				$client->{config}->{username}
					= $clientcfg{by_key}{username};
	
				$username{name} 
					= $clientcfg{by_key}{username};
			}
			# team
			if(exists $clientcfg{by_key}{team} &&
			   $clientcfg{by_key}{team} >= 0
			) {
				$client->{config}->{team} 
					= $clientcfg{by_key}{team};

				$team_list->{team_list}->{team}->{$clientcfg{by_key}{team}}->{number}
					= $clientcfg{by_key}{team};
				
				$username{team} 
					= $clientcfg{by_key}{team};
			}
			# machineid
			if($clientcfg{by_key}{machineid}) {
				$client->{config}->{machine_id} 
					= $clientcfg{by_key}{machineid};
			}
			# bigpackets=yes
			if($clientcfg{by_key}{bigpackets}) {
				$client->{config}->{big_wu} 
					= $clientcfg{by_key}{bigpackets};
			}
			else {
				$client->{config}->{big_wu}
					= 'no-pref';
			}
			# type=1
			if($clientcfg{by_key}{type}) {
				$client->{config}->{type}
					= $clientcfg{by_key}{type};

				if($client_type{$client->{config}->{type}}) {
					$client->{config}->{type_name}
						= $client_type{$client->{config}->{type}};
				}
			}
			else {
				$client->{config}->{type}
					= '0';
	
				if($client_type{$client->{config}->{type}}) {
					$client->{config}->{type_name}
						= $client_type{$client->{config}->{type}};
				}
			}
			# extra_parms=-smp 2 -verbosity 9 -forceasm
			if($clientcfg{by_key}{extra_parms}) {
				$client->{software}->{fah}->{arguments}
					= $clientcfg{by_key}{extra_parms};
			}
		
			# add client.cfg to the file list of this client
			$client->{files}->{client_cfg} = 'client-data/'.$client_name.'/client.cfg';
			$client_list->{client_list}->{client}->{"$client_name"}->{files}->{client_cfg} = $client->{files}->{client_cfg};
		
			# copy client.cfg.new to client.cfg 
			print "Copying ".basename($file)." to: $client->{files}->{client_cfg}\n" if($cfg{verbose});
			copy($file, $client->{files}->{client_cfg}) || die "Error: Cannot copy $file -> $client->{files}->{client_cfg} ($!)\n";
			chmod(0664, $client->{files}->{client_cfg});

			print "\n"  if($cfg{verbose});
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{client_cfg_new};
				delete $client->{files}->{client_cfg_new};
			}
			
			print "Error: Cannot read client.cfg file: $file\n";
			exit 1;
		}
	}
	
	# Other files possibly uploaded:
	# FAHlog.txt
	# FAHlog-Prev.txt  /* not used atm */
	# MyFolding.html
	
	# A concatenation of FAHlog.txt and FAHlog-Prev.txt will be saved in
	# FAHlog-Complete.txt
	my $fahlog_complete_file = $cfg{'client-data-dir'}.$client_name.'/FAHlog-Complete.txt';
	
	# FAHlog.txt
	my %fahlog = ();
	if($client->{files}->{fahlog_new}) {
		my $file = $client->{files}->{fahlog_new};
	
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			%fahlog = parse_fahlog($file);
			#print Dumper(\%fahlog);
			
			# add FAHlog.txt to the file list of this client
        	        $client->{files}->{fahlog} = 'client-data/'.$client_name.'/FAHlog.txt';
        	        $client_list->{client_list}->{client}->{"$client_name"}->{files}->{fahlog} = $client->{files}->{fahlog};
	
	                # copy FAHlog.txt.new to FAHlog.txt
		        copy($file, $client->{files}->{fahlog}) || die "Error: Cannot copy $file -> $client->{files}->{fahlog} ($!)\n";
        	        chmod(0664, $client->{files}->{fahlog});

			# arguments may be set in client.cfg which is parsed before FAHlog.txt
			# any commandline arguments in FAHlog.txt should be appended to this list
			if(exists $client->{software}                     &&
			   exists $client->{software}->{fah}              &&
			   exists $client->{software}->{fah}->{arguments}
			) {
			   	if(exists $fahlog{client}            &&
			  	   exists $fahlog{client}{arguments}
				) {
					my @clientcfg_arguments = parse_arguments($client->{software}->{fah}->{arguments});
					my @fahlog_arguments    = parse_arguments($fahlog{client}{arguments});

					my @arguments = @clientcfg_arguments;

					foreach my $argument (@fahlog_arguments) {
						if(!grep $argument, @clientcfg_arguments) {
							push @arguments, $argument;
						}
					}

					$fahlog{client}{arguments} = join ' ', @arguments;
				}
				else {
					$fahlog{client}{arguments} = $client->{software}->{fah}->{arguments};
				}
			}

			$client->{software}->{fah} = $fahlog{client};

			if($fahlog{client_shutdown} && !$fahlog{client_shutdown}{continued}) {
				$client_list->{client_list}->{client}->{$client_name}->{client_shutdown} = 1;

				$client->{client_shutdown} = 1;
			}

			# add FAHlog.txt to the file list of this client
			$client->{files}->{fahlog} = 'client-data/'.$client_name.'/FAHlog.txt';
			$client_list->{client_list}->{client}->{"$client_name"}->{files}->{fahlog} = $client->{files}->{fahlog};
		
			# copy FAHlog.txt.new to FAHlog.txt 
			print "Copying ".basename($file)." to: $client->{files}->{fahlog}\n" if($cfg{verbose});
			copy($file, $client->{files}->{fahlog}) || die "Error: Cannot copy $file -> $client->{files}->{fahlog} ($!)\n";
			chmod(0664, $client->{files}->{fahlog});

			print "\n" if($cfg{verbose});
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog_new};
				delete $client->{files}->{fahlog_new};
			}
			
			print "Error: Cannot read FAHlog.txt file: $file\n";
			exit 1;
		}
	}

	# FAHlog-Prev.txt
	if($client->{files}->{fahlog_prev_new}) {
		my $file = $client->{files}->{fahlog_prev_new};
	
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			# add FAHlog-Prev.txt to the file list of this client
        	        $client->{files}->{fahlog_prev} = 'client-data/'.$client_name.'/FAHlog-Prev.txt';
        	        $client_list->{client_list}->{client}->{"$client_name"}->{files}->{fahlog_prev} = $client->{files}->{fahlog_prev};
	
	                # copy FAHlog-Prev.txt.new to FAHlog-Prev.txt
			print "Copying ".basename($file)." to: $client->{files}->{fahlog_prev}\n" if($cfg{verbose});
		        copy($file, $client->{files}->{fahlog_prev}) || die "Error: Cannot copy $file -> $client->{files}->{fahlog_prev} ($!)\n";
        	        chmod(0664, $client->{files}->{fahlog_prev});

			print "\n" if($cfg{verbose});
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog_prev_new};
				delete $client->{files}->{fahlog_prev_new};
			}
		}
	}
	
	# FAHlog-Complete.txt 
	my %fahlog_complete = ();
	if($client->{files}->{fahlog_complete_new}) {
		my $file = $client->{files}->{fahlog_complete_new};

		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file && $client->{state} eq 'active') {
			print "Parsing FAHlog-Complete.txt file: $file\n" if($cfg{verbose});

			open(F, $file) || die "Cannot open log file: $file\n";
			my @lines = <F>;
			close F;

			#########################
			# Parse work attempts
			# 
			# [20:11:28] + Attempting to get work packet
			# [20:11:28] - Will indicate memory of 3966 MB
			# [20:11:28] - Connecting to assignment server
			# [20:11:28] Connecting to http://assign.stanford.edu:8080/
			# [20:11:29] Posted data.
			# [20:11:29] Initial: 40AB; - Successful: assigned to (171.64.65.56).
			# [20:11:29] + News From Folding@Home: Welcome to Folding@Home
			# [20:11:29] Loaded queue successfully.
			# [20:11:29] Connecting to http://171.64.65.56:8080/
			# [20:11:29] - Couldn't send HTTP request to server
			# [20:11:29]   (Got status 503)
			# [20:11:29] + Could not connect to Work Server
			# [20:11:29] - Attempt #6  to get work failed, and no other work to do.
			# Waiting before retry.
			
			if($lines[-1] =~ /^\s*Waiting before retry\./ && $lines[-2] =~ /^\[\d{2}:\d{2}:\d{2}\] \- (?:Error: )?(Attempt #\d+\s+to get work failed, and no other work to do\.)/) {
				my $message = $1;
				my $line    = $#lines;

				print "Unable to get work at line $line: $message\n" if($cfg{verbose});
				
				$client->{client_retry_wait}      = $message;
				$client->{client_retry_wait_line} = $line;
				
				$client_list->{client_list}->{client}->{$client_name}->{client_retry_wait}      = $message;
				$client_list->{client_list}->{client}->{$client_name}->{client_retry_wait_line} = $line;
			}
			# Folding@Home Client Shutdown
			elsif($lines[-1] =~ /^(Folding\@Home Client Shutdown)/) {
				my $message = $1;
				my $line    = $#lines + 1;

				print "Client shutdown at line $line: $lines[-1]" if($cfg{verbose});

				$client->{client_shutdown}      = $message;
				$client->{client_shutdown_line} = $line;

				$client_list->{client_list}->{client}->{$client_name}->{client_shutdown}      = $message;
				$client_list->{client_list}->{client}->{$client_name}->{client_shutdown_line} = $line;
			}
			# [19:45:22] EUE limit exceeded. Pausing 24 hours.
			elsif($lines[-1] =~ /^\[\d{2}:\d{2}:\d{2}\] (EUE limit exceeded\. Pausing 24 hours\.)/) {
				my $message = $1;
				my $line    = $#lines + 1;

				print "EUE limit exceeded at line $line: $message\n" if($cfg{verbose});

				$client->{client_eue_limit}      = $message;
				$client->{client_eue_limit_line} = $line;
				
				$client_list->{client_list}->{client}->{$client_name}->{client_eue_limit}      = $message;
				$client_list->{client_list}->{client}->{$client_name}->{client_eue_limit_line} = $line;
			}
			
			# Get the Time Per Frame info for the Current Work Unit
			if($client->{current_work_unit}->{project_number}                   &&
			   $client->{current_work_unit}->{project_number}        =~ /^\d+$/ &&
			   $client->{current_work_unit}->{folding}                          &&
			   $client->{current_work_unit}->{folding}->{run}        =~ /^\d+$/ &&
			   $client->{current_work_unit}->{folding}->{clone}      =~ /^\d+$/ &&
			   $client->{current_work_unit}->{folding}->{generation} =~ /^\d+$/ &&
			   $client->{current_work_unit}->{begin_timestamp}                  &&
			   $client->{current_index} >= 0 && $client->{current_index} <= 9
			) {
                	        my $project    = $client->{current_work_unit}->{project_number};
                	        my $run        = $client->{current_work_unit}->{folding}->{run};
                	        my $clone      = $client->{current_work_unit}->{folding}->{clone};
                	        my $generation = $client->{current_work_unit}->{folding}->{generation};
                	        my $begin      = $client->{current_work_unit}->{begin_timestamp};
				my $index      = sprintf("%02d", $client->{current_index});
				
				my ($min, $max, $avg, $cur, $tot, $frames, @steps) = get_tpf($project, $run, $clone, $generation, $begin, $index, $client->{last_update}, \@lines, %cfg);

				if(($min ne '' || ($min =~ /^\d+$/ && $min > 0)) &&
				   ($max ne '' || $max > 0) &&
				   ($avg ne '' || $avg > 0) &&
				   ($tot ne '' || $tot > 0) &&
				   ($frames > 0)
				) {
					$client->{current_work_unit}->{time_per_frame}->{minimum}        = $min;
					$client->{current_work_unit}->{time_per_frame}->{maximum}        = $max;
					$client->{current_work_unit}->{time_per_frame}->{average}        = $avg;
					$client->{current_work_unit}->{time_per_frame}->{last}           = $cur;
					$client->{current_work_unit}->{time_per_frame}->{total}          = $tot;
					$client->{current_work_unit}->{time_per_frame}->{frames}         = $frames;
				
					# Generate the javascript arrays used by jQuery Flot 
					my ($flot_data, $flot_data_time, $flot_data_line) = generate_flot_data(\@steps, %cfg);
				
					if($flot_data ne '' && $flot_data_time ne '' && $flot_data_line ne '') {
						$client->{current_work_unit}->{time_per_frame}->{flot_data}      = $flot_data;
						$client->{current_work_unit}->{time_per_frame}->{flot_data_time} = $flot_data_time;
						$client->{current_work_unit}->{time_per_frame}->{flot_data_line} = $flot_data_line;
					}
	
					$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{time_per_frame}
						= $client->{current_work_unit}->{time_per_frame};

					if($client->{current_work_unit}->{status} eq 'folding now') {
						# check for stalled/hung clients, no progress for the time of the last 2 frames.
						my $check_time = $steps[-1]->{epoch} + ($steps[-2]->{seconds} + $steps[-1]->{seconds});
						
						my $client_date = $qddata{executed_utc} ? $qddata{executed_utc} : '';

						# Fri Jul 17 07:07:10 UTC 2009
						if($client_date =~ /^\w+\s+(\w+)\s+(\d+)\s+(\d{2}:\d{2}:\d{2})\s+UTC\s+(\d{4})$/) {
							my $month = $1;
							my $day   = $2;
							my $time  = $3;
							my $year  = $4;

							$month = Date::Calc::Decode_Month($month);
							$month = '0'.$month if(length($month) == 1);

							my $client_time = UnixDate($year.'-'.$month.'-'.$day.' '.$time.' UTC', "%s");
						
							if($client_time > $check_time) {
								my $time_diff = time_format($client_time - $steps[-1]->{epoch}, 'long');

								my $message = "No progress for $time_diff";
								my $line    = $steps[-1]->{line};

								$client->{client_hanging}      = $message;
								$client->{client_hanging_line} = $line;

								$client_list->{client_list}->{client}->{$client_name}->{client_hanging}      = $message;
								$client_list->{client_list}->{client}->{$client_name}->{client_hanging_line} = $line;

								print "Client hanging: $message\n" if($cfg{verbose});
							}
						}

						# calculate the expected time of completion 
						#my $expected = get_expected(\@steps, %cfg, steps_avg => $avg);

						#print "Expected: $expected\n" if($cfg{debug});
					}
				}
			}

			# Get the Time Per Frame info for the last completed work unit
			
			if(0 == 1) {
			my $previous_index = previous_index($client->{current_index});
			foreach(0..9) {
				if($qddata{queue}{index}{$previous_index}{status} =~ /^(finished|ready for upload)$/) {
					# get the info
					if($qddata{queue}{index}{$previous_index}->{project_number}                   &&
					   $qddata{queue}{index}{$previous_index}->{project_number}        =~ /^\d+$/ &&
					   $qddata{queue}{index}{$previous_index}->{folding}                          &&
					   $qddata{queue}{index}{$previous_index}->{folding}->{run}        =~ /^\d+$/ &&
					   $qddata{queue}{index}{$previous_index}->{folding}->{clone}      =~ /^\d+$/ &&
					   $qddata{queue}{index}{$previous_index}->{folding}->{generation} =~ /^\d+$/ &&
					   $qddata{queue}{index}{$previous_index}->{begin_timestamp}                  &&
					   $client->{current_index} >= 0 && $client->{current_index} <= 9
					) {
		                	        my $project    = $qddata{queue}{index}{$previous_index}->{project_number};
		                	        my $run        = $qddata{queue}{index}{$previous_index}->{folding}->{run};
		                	        my $clone      = $qddata{queue}{index}{$previous_index}->{folding}->{clone};
		                	        my $generation = $qddata{queue}{index}{$previous_index}->{folding}->{generation};
		                	        my $begin      = $qddata{queue}{index}{$previous_index}->{begin_timestamp};
						my $index      = sprintf("%02d", $previous_index);
						
						my ($min, $max, $avg, $cur, $tot, $frames, $flot_data, $flot_data_time, $flot_data_line)
							= get_tpf($project, $run, $clone, $generation, $begin, $index, $client->{last_update}, \@lines, %cfg);
						
						if(($min ne '' || ($min =~ /^\d+$/ && $min > 0)) &&
						   ($max ne '' || $max > 0) &&
						   ($avg ne '' || $avg > 0) &&
						   ($tot ne '' || $tot > 0) &&
						   ($frames > 0)
						) {
						}
						
						last; # don't process any further queue indices
					}

					# also check one index further back which we may not have processed completely yet
					$previous_index = previous_index($previous_index);
				}
				else {
					if($qddata{queue}{index}{$previous_index}{status} eq 'deleted') {
						# get the info

						# also check one index further back which we may not have processed completely yet
						$previous_index = previous_index($previous_index);
					}
					else {
						last; # don't process any further queue indices
					}
				}
			}	
			}

			# add FAHlog-Complete.txt to the file list of this client
        	        $client->{files}->{fahlog_complete} = 'client-data/'.$client_name.'/FAHlog-Complete.txt';
        	        $client_list->{client_list}->{client}->{"$client_name"}->{files}->{fahlog_complete} = $client->{files}->{fahlog_complete};
	
	                # copy FAHlog-Complete.txt.new to FAHlog-Complete.txt
			print "Copying ".basename($file)." to: $client->{files}->{fahlog_complete}\n" if($cfg{verbose});
		        copy($file, $client->{files}->{fahlog_complete}) || die "Error: Cannot copy $file -> $client->{files}->{fahlog_complete} ($!)\n";
        	        chmod(0664, $client->{files}->{fahlog_complete});

			print "\n" if($cfg{verbose});
		}
		elsif(-r $file && $client->{state} ne 'active') {
			print "Not parsing FAHlog-Complete.txt file: $file\nClient is not active\n" if($cfg{verbose});

			# add FAHlog-Complete.txt to the file list of this client
        	        $client->{files}->{fahlog_complete} = 'client-data/'.$client_name.'/FAHlog-Complete.txt';
        	        $client_list->{client_list}->{client}->{"$client_name"}->{files}->{fahlog_complete} = $client->{files}->{fahlog_complete};
	
	                # copy FAHlog-Complete.txt.new to FAHlog-Complete.txt
			print "Copying ".basename($file)." to: $client->{files}->{fahlog_complete}\n" if($cfg{verbose});
		        copy($file, $client->{files}->{fahlog_complete}) || die "Error: Cannot copy $file -> $client->{files}->{fahlog_complete} ($!)\n";
        	        chmod(0664, $client->{files}->{fahlog_complete});

			print "\n" if($cfg{verbose});
		}
                else {
                        if(!-e $file) {
                                delete $client_list->{client_list}->{client}->{$client_name}->{files}->{fahlog_complete_new};
                                delete $client->{files}->{fahlog_complete_new};
                        }

                        print "Error: Cannot read FAHlog-Complete.txt file: $file\n";
                        exit 1;
                }
	}
	
	# MyFolding.html
	my %myfolding = ();
	if($client->{files}->{myfolding_new}) {
		my $file = $client->{files}->{myfolding_new};
	
		print "File to parse: $file\n" if($cfg{verbose});
		if(-r $file) {
			%myfolding = parse_myfolding($file);

			# username
			if($myfolding{username} && 
			   !$client->{config}->{username}) {
				$client->{config}->{username} 
					= $myfolding{username};
	
				$username{name} 
					= $myfolding{username};
			}
			
			# team
			if( (exists $myfolding{team}          && $myfolding{team}          >= 0) && 
			   !(exists $client->{config}->{team} && $client->{config}->{team} >= 0)
			) {
				$client->{config}->{team} 
					= $myfolding{team};

				$team_list->{team_list}->{team}->{$myfolding{team}}->{number}
					= $myfolding{team};
	
				$username{team} 
					= $myfolding{team};
			}

			# stats: userpage & teampage
			if($myfolding{statistics}) {
				$username{external_resources} = $myfolding{statistics};

				if($myfolding{statistics}->{teampage} && $client->{config}->{team} >= 0) {
					$team_list->{team_list}->{team}->{$client->{config}->{team}}->{external_resources}->{teampage} 
						= $myfolding{statistics}->{teampage};
				}
			}
			
			# add MyFolding.html to the file list of this client
        	        $client->{files}->{myfolding} = 'client-data/'.$client_name.'/MyFolding.html';
        	        $client_list->{client_list}->{client}->{"$client_name"}->{files}->{myfolding} = $client->{files}->{myfolding};
	
	                # copy MyFolding.html.new to MyFolding.html
			print "Copying ".basename($file)." to: $client->{files}->{myfolding}\n" if($cfg{verbose});
		        copy($file, $client->{files}->{myfolding}) || die "Error: Cannot copy $file -> $client->{files}->{myfolding} ($!)\n";
        	        chmod(0664, $client->{files}->{myfolding});

			print "\n" if($cfg{verbose});
		}
		else {
			if(!-e $file) {
				delete $client_list->{client_list}->{client}->{$client_name}->{files}->{myfolding_new};
				delete $client->{files}->{myfolding_new};
			}
			
			print "Error: Cannot read MyFolding.html file: $file\n";
			exit 1;
		}
	}

	# project specific stuff, only usefull when a project number is known
	if($client->{current_work_unit}->{project_number}) {
		my $project_number = $client->{current_work_unit}->{project_number};

		# Add the project to the known-projects.xml if it's not already listed
		if(!exists $known_projects->{known_projects}->{project}->{$project_number}) {
			if($client->{current_work_unit}->{project_name}) {
				$known_projects->{known_projects}->{project}->{$project_number}->{workunit_name}
					= $client->{current_work_unit}->{project_name};
			}

			$known_projects->{known_projects}->{project}->{$project_number}->{number}        = $project_number;
			$known_projects->{known_projects}->{project}->{$project_number}->{date_created}  = get_date();
		}
		
		# project-summary.xml
		if($projects->{project_summary}->{project}->{$project_number}) {	
			# core code
			if($projects->{project_summary}->{project}->{$project_number}->{code}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{code} 
					= $projects->{project_summary}->{project}->{$project_number}->{code};

				$client->{current_work_unit}->{code}
					= $projects->{project_summary}->{project}->{$project_number}->{code};
			}
		
			# project name
			if($projects->{project_summary}->{project}->{$project_number}->{name}) {
				$client_list->{client_list}->{client}->{$client_name}->{current_work_unit}->{name} 
					= $projects->{project_summary}->{project}->{$project_number}->{name};
				
				$client->{current_work_unit}->{project_name}
					= $projects->{project_summary}->{project}->{$project_number}->{name};
			}
		}
	
		# add the current workunit info of this client to the clients per projects list
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{name}
			= $client_name;
		
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{last_update}
			= $client->{last_update};
		
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{state}
			= $client->{state};
	
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{state_message}
			= $client->{state_message};
		
		if($client->{client_shutdown}) {
			$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{client_shutdown}
				= $client->{client_shutdown};
		}
	
		if($client->{hardware}) {
			$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{hardware}
				= $client->{hardware};
		}
		if($client->{operating_system}) {
			$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{operating_system}
				= $client->{operating_system};
		}
		if($client->{software}) {
			$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{software}
				= $client->{software};
		}
		
		$client_projects->{client_projects}->{project}->{$project_number}->{client_list}->{client}->{$client_name}->{current_work_unit}
			= $client_list->{client_list}->{client}->{"$client_name"}->{current_work_unit};
	}
	
	$client_list->{client_list}->{client}->{"$client_name"}->{config} = $client->{config} if($client->{config});
	$client_list->{client_list}->{client}->{"$client_name"}->{stats}  = $client->{stats}  if($client->{stats});
	
	# add username element to username_list
	if($username{name} && (exists $username{team} && $username{team} >= 0)) {
		print "Updating Stanford user- & teampage URLs.\n" if($cfg{verbose});
		update_stanford_urls(%username);
		print "\n" if($cfg{verbose});
	}

	#print Dumper($client);
	
	# save the client.xml file
	my $client_xml = $client->{files}->{client_xml};
	
	print "Saving updated client information to: $client_xml\n" if($cfg{verbose});
	XMLout( $client,
	        KeyAttr    => { client => 'name',  server => 'number', fci => 'version' },
	        OutputFile => $client_xml,
	        RootName   => 'client',
		NoAttr     => 1,
	      ) || die $@;

	if($cfg{verbose}) {
		print "\n";
		print "Successfully updated the client XML file for: $client_name\n";
	}

	# update fci summary with number of clients
	$fci_summary->{fci_summary}->{clients} += 1;

	$cfg{'write-xml'} = 1;

	my $client_processing_stop = new Benchmark;
	my $client_processing_diff = timediff($client_processing_stop, $client_processing_start);
	print "Client processed in: ". timestr($client_processing_diff) ."\n" if($cfg{verbose});

	print "\n\n" if($cfg{verbose});
} # parsed per client uploaded files

my $client_processing_stop = new Benchmark;
my $client_processing_diff = timediff($client_processing_stop, $client_processing_start);
print "Client processing completed in: ". timestr($client_processing_diff) ."\n\n" if($cfg{verbose});


# Add the project data from project-summary.xml to 
# the corresponding client-projects.xml element

$time_start = new Benchmark;

print "Adding data from project summary to client projects list.\n" if($cfg{verbose});

foreach my $project (keys %{$client_projects->{client_projects}->{project}}) {
	if($projects->{project_summary}->{project}->{$project}) {
		$client_projects->{client_projects}->{project}->{$project}->{project_summary} 
			= $projects->{project_summary}->{project}->{$project};
	}

	# update fci summary with number of client projects
	$fci_summary->{fci_summary}->{client_projects} += 1;
}

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);
print "Completed in: ". timestr($time_diff) ."\n\n" if($cfg{verbose});


# Add the project data from project-summary.xml to 
# the corresponding known-projects.xml element

$time_start = new Benchmark;

print "Adding data from project summary to known projects list.\n" if($cfg{verbose});
foreach my $project (keys %{$known_projects->{known_projects}->{project}}) {
	if($projects->{project_summary}->{project}->{$project}) {
		$known_projects->{known_projects}->{project}->{$project}->{project_summary} 
			= $projects->{project_summary}->{project}->{$project};
	}

	if($jmol_projects->{fah_projects}->{fah_proj}->{$project} &&
	   $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{file} &&
	   $jmol_projects->{fah_projects}->{fah_proj}->{$project}->{file} eq 'y'
	) {
		print "Adding URL to Jmol webpage to project: $project\n" if($cfg{verbose});
		$known_projects->{known_projects}->{project}->{$project}->{jmol_url}
			= "http://jmol.sourceforge.net/fah/?project=$project";
	}

	if(exists $known_projects->{known_projects}->{project}->{$project}->{project_summary}->{status}->{active} &&
	   $known_projects->{known_projects}->{project}->{$project}->{project_summary}->{status}->{active} eq 'false'
	) {
		# The project is not currently listed on the psummary page
		# If the project does not have any clients running it, consider it inactive

		if(exists $client_projects->{client_projects}->{project}->{$project} &&
		   exists $client_projects->{client_projects}->{project}->{$project}->{client_list}
		) {
			if(keys %{$client_projects->{client_projects}->{project}->{$project}->{client_list}} == 0) {
				$known_projects->{known_projects}->{project}->{$project}->{state} = 'inactive';
			}
			else {
				$known_projects->{known_projects}->{project}->{$project}->{state} = 'active';
			}
		}
		else {
			$known_projects->{known_projects}->{project}->{$project}->{state} = 'inactive';
		}
	}
	else {
		if(exists $known_projects->{known_projects}->{project}->{$project}->{project_summary}->{status}->{active} &&
		   $known_projects->{known_projects}->{project}->{$project}->{project_summary}->{status}->{active} eq 'true'
		) {
			$known_projects->{known_projects}->{project}->{$project}->{state} = 'active';
		}
		else {
			$known_projects->{known_projects}->{project}->{$project}->{state} = 'inactive';
		}
	}

	# Update the project data from known-projects.xml in the project specific XML file
	my $known_project_file = $cfg{'xml-data-dir'} .'known-project-'. $project .'.xml';

	if(-r $known_project_file) {
		my $known_project = read_xml( %cfg,
					      Name => 'known-project',
					      File => $known_project_file
					    );
	
		$known_project->{known_project}->{project} = $known_projects->{known_projects}->{project}->{$project};
	
		if(exists $client_projects->{client_projects}->{project}->{$project} &&
		   exists $client_projects->{client_projects}->{project}->{$project}->{client_list}
		) {
			my %client_list = %{$client_projects->{client_projects}->{project}->{$project}->{client_list}};

			$known_project->{known_project}->{client_list} = \%client_list;
		}
		elsif(exists $known_project->{known_project}->{client_list}) {
			# Remove the existing client list if we don't have a client list for this project anymore
			delete $known_project->{known_project}->{client_list};
		}
	
        	my ($success, $reason) = write_xml( %cfg,
        	                                    Name => 'known-project',
        	                                    File => $known_project_file,
        	                                    XML  => $known_project
        	                                  );
        	if(!$success) {
		        print $reason;
			exit 1;
		}
	}

	# update fci summary with number of client projects
	$fci_summary->{fci_summary}->{known_projects} += 1;
}

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);
print "Completed in: ". timestr($time_diff) ."\n\n" if($cfg{verbose});


# add the EXTREME Overclocking data to each user
#
# The User Summary URL is in the form:
# http://folding.extremeoverclocking.com/user_summary.php?s=&u=$USERID
#
# The UserID can be found in the users XML file in eoc-data/

$time_start = new Benchmark;

print "Parsing EXTREME Overclocking Folding\@Home Stats data.\n" if($cfg{verbose});

foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
	my $username = $username_list->{username_list}->{username}->{$userid}->{name};
	my $team     = $username_list->{username_list}->{username}->{$userid}->{team};

	# Reset the username & team state
	$username_list->{username_list}->{username}->{$userid}->{state} = "unknown";
	$team_list->{team_list}->{team}->{"$team"}->{state} 		= "unknown";

	my $file = $cfg{'eoc-data-dir'}.$username.'-'.$team.'.xml';
	
	print " Looking for XML file for user: $username team: $team ... " if($cfg{verbose});
	
	if(-e $file && -r $file) {
		if($cfg{verbose}) {
			print "Found.\n";
			print " File to parse: $file\n";
		}
		my %eoc_user_summary = parse_eoc_user_summary($file, %cfg);
		
		# construct the URL to the EOC F@H Stats User Summary page
		if(exists $eoc_user_summary{user}->{UserID} &&
		   $eoc_user_summary{user}->{UserID} =~ /^\d+$/
		) {
			my $url  = "http://folding.extremeoverclocking.com/user_summary.php?s=&u=";
			   $url .= $eoc_user_summary{user}->{UserID};

			$username_list->{username_list}->{username}->{$userid}->{external_resources}->{eoc_user_summary}
				= $url;
		}

		# construct the URL to the EOC F@H Stats Team Summary page
		if(exists $eoc_user_summary{team}->{TeamID} &&
		   $eoc_user_summary{team}->{TeamID} =~ /^\d+$/
		) {
			my $team = $eoc_user_summary{team}->{TeamID};
		
			my $url  = "http://folding.extremeoverclocking.com/team_summary.php?s=&t=";
			   $url .= $team;

			$team_list->{team_list}->{team}->{$team}->{external_resources}->{eoc_team_summary}
				= $url;
		}

		# extract the Points 24 Hour Average for this user
		if(exists $eoc_user_summary{user}->{Points_24hr_Avg} &&
		   $eoc_user_summary{user}->{Points_24hr_Avg} =~ /^\d+(\.\d+)?$/
		) {
			$username_list->{username_list}->{username}->{$userid}->{points_24hr_avg}
				= $eoc_user_summary{user}->{Points_24hr_Avg};

			# Users with a 24 hour average of 0 are considered inactive
			if($eoc_user_summary{user}->{Points_24hr_Avg} == 0) {
				$username_list->{username_list}->{username}->{$userid}->{state}
					= "inactive";
			}
			else {
				$username_list->{username_list}->{username}->{$userid}->{state}
					= "active";
			}
		}

		# extract Points Per Day average for this users team
		if(exists $eoc_user_summary{team}->{Points_24hr_Avg} &&
		   $eoc_user_summary{team}->{Points_24hr_Avg} =~ /^\d+(\.\d+)?$/
		) {
			my $team = $username_list->{username_list}->{username}->{$userid}->{team};
			
			$team_list->{team_list}->{team}->{"$team"}->{points_24hr_avg}
				= $eoc_user_summary{team}->{Points_24hr_Avg};

			# Users with a 24 hour average of 0 are considered inactive
			if($eoc_user_summary{team}->{Points_24hr_Avg} == 0) {
				$team_list->{team_list}->{team}->{"$team"}->{state}
					= "inactive";
			}
			else {
				$team_list->{team_list}->{team}->{"$team"}->{state}
					= "active";
			}
		}
	}
	else {
		if(-e $file && !-r $file) {
			print "Found, but unreadable!\n" if($cfg{verbose});
		}
		else {
			print "Not Found.\n" if($cfg{verbose});
		}
	}
	print "\n" if($cfg{verbose});
}

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);
print "Completed in: ". timestr($time_diff) ."\n\n" if($cfg{verbose});


# add the kakaostats.com user summary URL to each user
#
# The User Summary URL is in the form:
# http://kakaostats.com/usum.php?u=$USERID
#
# The UserID can be found in the fahstats XML file in xml-data/

if(-r $cfg{'fahstats-file'}) {
	$time_start = new Benchmark;

	print "Adding kakaostats.com User Summary URLs.\n\n" if($cfg{verbose});

	my $fahstats = XMLin( $cfg{'fahstats-file'},
		              KeyAttr    => { username => 'userid' },
			      KeepRoot   => 1,
			      ForceArray => [ 'username' ]
	                    ) || die "Error: Cannot read fahstats XML file! ($!)";

	foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
		if(exists $fahstats->{fahstats}->{username}->{$userid}->{fahstats_userid} &&
		   $fahstats->{fahstats}->{username}->{$userid}->{fahstats_userid} =~ /^\d+$/
		) {
			my $url  = "http://kakaostats.com/usum.php?u=";
			   $url .= $fahstats->{fahstats}->{username}->{$userid}->{fahstats_userid};
	
			$username_list->{username_list}->{username}->{$userid}->{external_resources}->{fahstats_user_summary}
				= $url;

		}
		
		if(exists $fahstats->{fahstats}->{username}->{$userid}->{team} &&
		   $fahstats->{fahstats}->{username}->{$userid}->{team} =~ /^\d+$/
		) {
			my $team = $fahstats->{fahstats}->{username}->{$userid}->{team};
		
			my $url  = "http://kakaostats.com/tsum.php?t=";
			   $url .= $team;
	
			$team_list->{team_list}->{team}->{$team}->{external_resources}->{fahstats_team_summary}
				= $url;

		}
	}

	$time_stop = new Benchmark;
	$time_diff = timediff($time_stop, $time_start);
	print "Completed in: ". timestr($time_diff) ."\n\n" if($cfg{verbose});
}

# add the Stanford userpage URL to each user
#
# The Stanford userpage URL is in the form:
# http://fah-web.stanford.edu/cgi-bin/main.py?qtype=userpage&username=$USERNAME&teamnum=$TEAMNUMBER
#
# The URL can be found in the MyFolding.html file, 

$time_start = new Benchmark;

print "Updating Stanford user- & teampage URLs.\n" if($cfg{verbose});
foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
	update_stanford_urls(%{$username_list->{username_list}->{username}->{$userid}});
}

$time_stop = new Benchmark;
$time_diff = timediff($time_stop, $time_start);
print "Completed in: ". timestr($time_diff) ."\n\n" if($cfg{verbose});


# update team_list with info from the teamstats file
#   update username_list with info from the teamstats file

my @no_teamstats = ();
foreach my $team (keys %{$team_list->{team_list}->{team}}) {
	my $teamstats_file = $cfg{'stanford-data-dir'}.'team'.$team.'.txt';
	
	print "Looking for stats file for team $team ... " if($cfg{verbose});
	if(-r $teamstats_file) {
		print "Found.\n" if($cfg{verbose});
		
		print "File to parse: $teamstats_file\n" if($cfg{verbose});
		my %team_stats = parse_teamstats($teamstats_file, $team);

		# name
		$team_list->{team_list}->{team}->{$team}->{name} 
			= $team_stats{team}->{name};
		
		# credits
		$team_list->{team_list}->{team}->{$team}->{score}
			= $team_stats{team}->{score};
		
		# work units
		$team_list->{team_list}->{team}->{$team}->{wu}
			= $team_stats{team}->{wu};
		
		# rank
		$team_list->{team_list}->{team}->{$team}->{rank}
			= $team_stats{team}->{rank};
		
		# last_update
		$team_list->{team_list}->{team}->{$team}->{last_update}
			= $team_stats{last_update};

		#print Dumper(\%teamstats);
		#print Dumper($username_list);

		my $users = 0;
		foreach my $team_rank (keys %{$team_stats{rank_entry}}) {
			# if the users in the team stats file is one of our fci users
			# update his info
			my $name = $team_stats{rank_entry}->{$team_rank}->{name};

			foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
				my $uname = $username_list->{username_list}->{username}->{$userid}->{name};
				my $uteam = $username_list->{username_list}->{username}->{$userid}->{team};

				if( ($team eq $uteam) && ($name eq $uname) ) {
					print "Parsing entry for user: $name team: $team\n" if($cfg{verbose});
					
					# team_rank, rank inside the team
					$username_list->{username_list}->{username}->{$userid}->{team_rank}
						= $team_rank;

					# rank, rank inside the whole project
					$username_list->{username_list}->{username}->{$userid}->{rank}
						= $team_stats{rank_entry}->{$team_rank}->{rank};

					# credit
					$username_list->{username_list}->{username}->{$userid}->{score}
						= $team_stats{rank_entry}->{$team_rank}->{credit};

					# work units
					$username_list->{username_list}->{username}->{$userid}->{wu}
						= $team_stats{rank_entry}->{$team_rank}->{total};
					
					
					# update fci summary with number of usernames
					$fci_summary->{fci_summary}->{usernames} += 1;
				}
			}
			
			$users++;
		}
		if($users > 0) {
			# members
	 		$team_list->{team_list}->{team}->{$team}->{members} = $users;
		}
	}
	else {
		# in case no teamstats file exists, fall back to daily_*_summary.txt files
		print "Not Found.\n" if($cfg{verbose});
		
		push @no_teamstats, $team;
	}
	
	# update fci summary with number of client projects
	$fci_summary->{fci_summary}->{teams} += 1;
	
	print "\n" if($cfg{verbose});
}


# update team_list with the info from the daily_team_summary.txt file
#   update username_list with the info from the daily_user_summary.txt file
#
# For one or more team, the teamstats where not available
# get the info from daily_team_summary.txt and daily_user_summary.txt

if($#no_teamstats >= 0) {
	print "Looking for daily_team_summary.txt file ... " if($cfg{verbose});
	if(-r $cfg{'team-summary-file'}) {
		print "Found.\n" if($cfg{verbose});
		
		my $i = 0;
		open(F, $cfg{'team-summary-file'}) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			#Tue Aug 30 12:00:00 PDT 2005
			#team    teamname        score   wu
			#446     Google  359611292       4925020
			if(/^(\d+)\t(.*?)\t(\d+)\t(\d+)$/) {
				my $team     = $1;
				my $teamname = $2;
				my $score    = $3;
				my $wu       = $4;

				# The Default and the Google teams both have rank 0
				# as they are special teams, the others have normal ranks
				if($team ne '0' && $team ne '446') {
					$i++;
				}
				
				foreach(@no_teamstats) {
					if($team eq $_) {
						print "Parsing entry for team: $_ name: $teamname\n" if($cfg{verbose});
						
						# name
						$team_list->{team_list}->{team}->{$team}->{name}
							= $teamname;
							
						# credits
						$team_list->{team_list}->{team}->{$team}->{score}
							= $score;
						
						# work units
						$team_list->{team_list}->{team}->{$team}->{wu}
							= $wu;
						
						# rank
						$team_list->{team_list}->{team}->{$team}->{rank}
							= $i;
						
						# members
						# remove the member count since we can only determin that from
						# the teamstats or daily_user_summary.txt
						if($team_list->{team_list}->{team}->{$team}->{members}) {
					 		delete $team_list->{team_list}->{team}->{$team}->{members};
						}
					}
				}
			}
		}
		close F;
	}
	else {
		print "Not Found.\n" if($cfg{verbose});
	}
	print "\n" if($cfg{verbose});

	print "Looking for daily_user_summary.txt file ... " if($cfg{verbose});
	if(-r $cfg{'user-summary-file'}) {
		print "Found.\n" if($cfg{verbose});
		
		open(F, $cfg{'user-summary-file'}) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			#Tue Mar 28 12:00:00 PST 2006
			#name    round(credit)   total   team
			#anonymous       66307146        911359  0
			if(/^(.*?)\t(\d+)\t(\d+)\t(\d+)$/) {
				my $name  = $1;
				my $score = $2;
				my $wu    = $3;
				my $team  = $4;

				foreach my $userid (keys %{$username_list->{username_list}->{username}}) {
					my $uname = $username_list->{username_list}->{username}->{$userid}->{name};
					my $uteam = $username_list->{username_list}->{username}->{$userid}->{team};
					
					my $has_teamstats = 1;
					foreach(@no_teamstats) {
						if($team eq $_) {
							$has_teamstats = 0;
						}
					}

					if( ($team eq $uteam) && ($name eq $uname) && !$has_teamstats ) {
						print "Parsing entry for user: $name team: $team\n" if($cfg{verbose});

						# credit
						$username_list->{username_list}->{username}->{$userid}->{score}
							= $score;
		
						# work units
						$username_list->{username_list}->{username}->{$userid}->{wu}
							= $wu;
						
						
						# update fci summary with number of usernames
						$fci_summary->{fci_summary}->{usernames} += 1;
					}
				}
			}
		}
	}
	else {
		print "Not Found.\n" if($cfg{verbose});
	}
	print "\n" if($cfg{verbose});
}

print "\n" if($cfg{verbose});

if($cfg{'write-xml'}) {
	#print Dumper($client_list);
	
	# save the client-list.xml file
	XMLout( $client_list,
		OutputFile => $cfg{'client-list-file'},
		KeyAttr    => { client => 'name' },
		NoAttr     => 1,
		KeepRoot   => 1
	      ) || die $@;
	
	if($cfg{verbose}) {
		print "Successfully updated the client list XML file.\n";
	}
	
	# save the team-list.xml file
	XMLout( $team_list,
		OutputFile => $cfg{'team-list-file'},
		KeyAttr    => 'number',
		NoAttr     => 1,
		KeepRoot   => 1
	      );
	
	if($cfg{verbose}) {
		print "Successfully updated the team list XML file.\n";
	}
		
	# save the username-list.xml file
	XMLout( $username_list,
		OutputFile => $cfg{'username-list-file'},
		KeyAttr    => 'userid',
		NoAttr     => 1,
		KeepRoot   => 1
	      ) || die $@;
	
	if($cfg{verbose}) {
		print "Successfully updated the username list XML file.\n";
	}
	
	# save the client-projects.xml file
	XMLout( $client_projects,
	        OutputFile => $cfg{'client-projects-file'},
		KeyAttr    => { project => 'number', client => 'name' },
		NoAttr     => 1,
		KeepRoot   => 1
	      ) || die $@;
	
	if($cfg{verbose}) {
		print "Successfully updated the client projects XML file.\n";
	}
	
	# save the known-projects.xml file
	XMLout( $known_projects,
	        OutputFile => $cfg{'known-projects-file'},
		KeyAttr    => { project => 'number', image => 'type' },
		NoAttr     => 1,
		KeepRoot   => 1
	      ) || die $@;
	
	if($cfg{verbose}) {
		print "Successfully updated the client projects XML file.\n";
	}
}
else {
	print "Warning: No client data processed, not updating XML files.\n";
}

# log time of completion in the FCI summary
$fci_summary->{fci_summary}->{last_update}->{end_time} = get_date();

# save the fci-summary.xml file
XMLout( $fci_summary,
        OutputFile => $cfg{'fci-summary-file'},
	NoAttr     => 1,
	KeepRoot   => 1
      ) || die $@;

if($cfg{verbose}) {
	print "Successfully updated the FCI summary XML file.\n";
}
										      

################################[ Subroutines ]################################


##############
#
#     Routine: read_xml()
#  Parameters: %args
#     Returns: a hashreference of the XML structure
# Description: read an XML file while enforcing the correct structure
#

sub read_xml {
	my %args = @_;
	
	if(!$args{Name}) {
		warn "FCI read_xml: Unknown name!\n";	
		return undef;
	}

	if($args{Name} =~ /^(client|qd|team-stats|known-project)$/) {
		if(!$args{File}) {
			warn "FCI read_xml: No file specified! ($args{Name})\n";
			return undef;
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			warn "FCI read_xml: Unknown file: $args{Name}\n";
			return undef;	
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File}) {
		warn "FCI read_xml: File does not exist: $args{File}\n";
		return undef;
	}
	if(-e $args{File} && !-r $args{File}) {
		warn "FCI read_xml: Cannot read file: $args{File}\n";
		return undef;
	}
	if(!-f $args{File}) {
		warn "FCI read_xml: Not a file: $args{File}\n";
		return undef;
	}
	
	if($args{File} && -r $args{File} && -f $args{File}) {
		if($args{Name} eq 'qd-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' },
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			return XMLin( $args{File},
			              KeyAttr    => 'team_rank',
				      ForceArray => [ 'rank_entry' ],
				      KeepRoot   => 1,
				    ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team_stats => 'team' }, 
				      ForceArray => [ 'team_stats' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			return XMLin( $args{File},
			              KeyAttr    => { file => 'id' }, 
				      ForceArray => [ 'file' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			return XMLin( $args{File},
			              KeyAttr    => { project  => 'number', client => 'name' }, 
				      ForceArray => [ 'project', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			return XMLin( $args{File},
			              KeyAttr    => { image  => 'type',  setting  => 'id', 
				                      option => 'value', location => 'id' },
				      ForceArray => [ 'image', 'setting', 'option', 'location' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { project => 'number', image => 'type' },
				      ForceArray => [ 'project', 'image' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'known-project') {
			return XMLin( $args{File},
				      KeyAttr    => { trajectory => 'id', image => 'type', client => 'name' },
				      ForceArray => [ 'trajectory', 'image', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			return XMLin( $args{File},
			              KeyAttr    => { application => 'name', site => 'name' },
				      ForceArray => [ 'application', 'site' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			return XMLin( $args{File},
			              KeyAttr    => { path => 'name' },
				      ForceArray => [ 'path' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-users') {
			return XMLin( $args{File},
			              KeyAttr    => { user => 'name', client => 'name' },
				      ForceArray => [ 'user', 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-clients') {
			return XMLin( $args{File},
			              KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'site-access-control') {
			return XMLin( $args{File},
			              KeyAttr    => { acl => 'name', client => 'name', user => 'name' },
				      ForceArray => [ 'acl', 'client', 'user' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'qd') {
			return XMLin( $args{File},
				      KeyAttr    => { index => 'number', server => 'number' },
				      ForceArray => [ 'index', 'server' ]
				    ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name', 'server' => 'number' },
				      ForceArray => [ 'server' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			return XMLin( $args{File},
				      KeyAttr    => { client => 'name' },
				      ForceArray => [ 'client' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			return XMLin( $args{File},
			              KeyAttr    => { username => 'userid' },
				      ForceArray => [ 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			return XMLin( $args{File},
			              KeyAttr    => { team => 'number', username => 'userid' },
				      ForceArray => [ 'team', 'username' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			return XMLin( $args{File},
			              KeyAttr    => { project => 'number' },
				      ForceArray => [ 'project' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
		elsif($args{Name} eq 'fci-summary') {
			return XMLin( $args{File},
				      KeepRoot   => 1
				    ) || die $@;
		}
                elsif($args{Name} eq 'jmol-missing-projects') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'project' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
		elsif($args{Name} eq 'jmol-projects') {
			return XMLin( $args{File},
				      KeyAttr    => { fah_proj => 'number' },
				      ForceArray => [ 'fah_proj' ],
				      KeepRoot   => 1
				    ) || die $@;
		}
                elsif($args{Name} eq 'queue-graphs') {
                        return XMLin( $args{File},
                                      ForceArray => [ 'client' ],
                                      KeepRoot   => 1
                                    ) || die $@;
                }
		else {
			# unknown file
			return undef;
		}
	}

	return undef;
}


##############
#
#     Routine: write_xml()
#  Parameters: name of the XML file 
#              xml structure in a hash reference
#              %args
#     Returns: a hashreference of the XML structure
#     Returns: 1 on sucess
#              0 and string with reasons on errors
# Description: write an XML file while enforcing the correct structure
#

sub write_xml {
	my %args = @_;
	
	my ($success, $reason) = (1, '');
	
	if(!$args{Name}) {
		return (0, "Unknown name!\n");
	}
	
	if($args{Name} =~ /^(client|qd|known-project)$/) {
		if(!$args{File}) {
			return (0, "No file specified! ($args{Name})\n");
		}
	}
	else {
		if(!$args{$args{Name}.'-file'} ) {
			return (0, "Unknown file: $args{Name}\n");
		}
		if(!$args{File}) {
			$args{File} = $args{$args{Name}.'-file'};
		}
	}
	
	if(!-e $args{File} && !-w dirname($args{File}) ) {
		return (0, "Cannot write directory for: $args{Name}\n");
	}
	if(-e $args{File} && !-w $args{File}) {
		return (0, "Cannot write file: $args{Name}\n");
	}
	if(-e $args{File} && !-f $args{File}) {
		return (0, "Not a file: ($args{Name})\n");
	}
	
	
	if($success) {
		if($args{Name} eq 'qd-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'team-stats') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => 'team_rank',
				NoAttr     => 1,
				KeepRoot   => 1,
		              ) || die $@;
		}
		elsif($args{Name} eq 'teamstats-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team_stats => 'team' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'stanford-files') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { file => 'id' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project  => 'number', client => 'name' }, 
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'project-images') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { image  => 'type',  setting  => 'id', 
				                option => 'value', location => 'id' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'known-projects') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { project => 'number', image => 'type' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'known-project') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { trajectory => 'id', image => 'type', client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
		              ) || die $@;
		}
		elsif($args{Name} eq 'site-defaults') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-config') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { application => 'name', site => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-permissions') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { path => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-users') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { user => 'name', client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-clients') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'site-access-control') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
			        KeyAttr    => { acl => 'name', client => 'name', user => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name', server => 'number', 'fci' => 'version' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'client-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'username-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'team-list') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { team => 'number', username => 'userid' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'xml-project-summary') {
			XMLout( $args{XML},
			        OutputFile => $args{File},
				KeyAttr    => { project => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-missing-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		elsif($args{Name} eq 'jmol-projects') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { fah_proj => 'number' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
                elsif($args{Name} eq 'queue-graphs') {
			XMLout( $args{XML},
				OutputFile => $args{File},
				KeyAttr    => { client => 'name' },
				NoAttr     => 1,
				KeepRoot   => 1
			      ) || die $@;
		}
		else {
			# unknown file
			$success = 0;
			$reason .= "Unknown file: $args{Name}\n";
		}
	}

	return ($success, $reason);
}


##############
#
#     Routine: delta_format()
#  Parameters: Date::Manip::DateCalc($date1, $date2, {}, 1) delta
#     Returns: time in semi human readable format: 1d2h3m4s
# Description: convert delta to readable format
#

sub delta_format {
	my $delta = shift;
	my $type  = shift;

	#+0:0:0:0:1:55:43
        if($delta =~ /^(\+|\-)(\d+):(\d+):(\d+):(\d+):(\d+):(\d+):(\d+)$/) {
                #+Y:+M:+W:+D:+H:+MN:+S
		my $sign    = $1;
                my $years   = $2;
                my $months  = $3;
                my $weeks   = $4;
                my $days    = $5;
                my $hours   = $6;
                my $minutes = $7;
                my $seconds = $8;

                my $time  = '';
		   $time .= $years  ."y"   if($years   > 0);
		   $time .= $months ."M"   if($months  > 0);
		   $time .= $weeks  ."w"   if($weeks   > 0);
		   $time .= $days   ."d"   if($days    > 0);
		   $time .= $hours  ."h"   if($hours   > 0);
		   $time .= $minutes."m"   if($minutes > 0);
		   $time .= $seconds."s"   if($seconds > 0);
		   $time .= ' ago' 	   if($sign eq '-');

		if($type && $type eq 'long') {
			my $long = '';

			if($time =~ /(\d+)y/) {
				my $n = $1;
				$long .= "$n year";
				$long =~ s/year/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)M/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n month";
				$long =~ s/month/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)w/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n week";
				$long =~ s/week/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)d/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n day";
				$long =~ s/day/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)h/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n hour";
				$long =~ s/hour/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)m/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n minute";
				$long =~ s/minute/$&s/ if($n != 1);
			}
			if($time =~ /(\d+)s/) {
				my $n = $1;
				$long .= " " if($long ne '');
				$long .= "$n second";
				$long =~ s/second/$&s/ if($n != 1);
			}
			if($time =~ /ago$/) {
				$long .= " $&";
			}

			$delta = $long;
		}
		else {
			$delta = $time;
		}
	}

	return $delta;
}

##############
#
#     Routine: time_format()
#  Parameters: time in seconds
#     Returns: time in semi human readable format: 1d2h3m4s
# Description: convert time in seconds to readable format
#

sub time_format {
        my $value = shift;
	my $type  = shift;
	   $type  = '' if(!$type);

        if($value && ($value >= 0)) {
                # convert time in seconds to human readable format
                # e.g. 65 seconds == 1m5s

                my $date1 = ParseDate("today");
                my $date2 = DateCalc($date1, "+ $value seconds");

                my $delta = DateCalc($date1, $date2, {}, 1);
                #print "Delta: $delta\n";

		$value = delta_format($delta, $type);
        }

        return $value;
}

##############
#
#     Routine: get_date()
#  Parameters: optionally a hash with the keys date and/or time set to something
#     Returns: the date in yyyy-mm-dd format, if only date is set
#              the time in hh:mm:ss format, if only time is set
#              the full date in yyyy-mm-dd hh:mm:ss format, if neither date and
#              time are set
#              both the date and the time, if they're both set
# Description: Get the contents and other info of uploaded data
#

sub get_date {
        my %option = @_;

        my ($year,$month,$day,$hour,$min,$sec) = Date::Calc::Today_and_Now();

        $month = '0'.$month if(length($month) == 1);
        $day   = '0'.$day   if(length($day)   == 1);
        $hour  = '0'.$hour  if(length($hour)  == 1);
        $min   = '0'.$min   if(length($min)   == 1);
        $sec   = '0'.$sec   if(length($sec)   == 1);

        my $date = "$year-$month-$day";
        my $time = "$hour:$min:$sec";

        if(!$option{date} && !$option{time}) {
                return "$date $time";
        }
        elsif($option{time} && !$option{date}) {
                return $time;
        }
        elsif($option{date} && !$option{time}) {
                return $date;
        }
        else {
                return ($date, $time);
        }
}

sub parse_arguments {
	my $arguments = shift;
	my @arguments = split /\s+/, $arguments;

	my @args = ();

	my $i = 0;
	foreach my $argument (@arguments) {
		if($argument eq '-delete'    ||
		   $argument eq '-send'      ||
		   $argument eq '-verbosity' ||
		   $argument eq '-smp'       ||
		   $argument eq '-gpu'
		) {
			if($arguments[$i+1] =~ /^\d+$/) {
				$argument .= ' '. $arguments[$i+1];
			}

			push @args, $argument;
		}
		elsif($argument eq '-forcegpu') {
			if($arguments[$i+1] =~ /^\w+$/) {
				$argument .= ' '. $arguments[$i+1];
			}

			push @args, $argument;
		}
		elsif($argument =~ /^-[a-zA-Z]+[a-zA-Z0-9]*?$/) {
			push @args, $argument;
		}

		$i++;
	}

	return @args;
}

sub parse_qddata {
	my $file = $_[0];
		
	print "Parsing qd-data.xml file: $file\n" if($cfg{verbose});
	
	my %info;
	if(-r $file) {
		my $qd = XMLin( $file,
				KeyAttr    => 'number',
				ForceArray => [ 'index', 'server' ]
			      ) || die $@;

		%info = %{$qd};
	}

	return %info;
}

sub parse_unitinfo {
	my $file = $_[0];

	print "Parsing unitinfo.txt file: $file\n" if($cfg{verbose});

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			# strip windows carriage return character
			if(s/\r//g) {}
			
			#Name: p1134_RIBO_FSpeptide_EXT
			if(/^Name: (.*)$/) {
				$info{project_name} = $1;
				
				if($info{project_name} =~ /^p(\d*)/) {
					$info{project_number} = $1;
				}
			}
			#Download time: January 31 07:33:58
			if(/^Download time: (.*)$/) {
				$info{download} = $1;
			}
			#Due time: February 1 22:44:58
			if(/^Due time: (.*)$/) {
				$info{due} = $1;
			}
			#Progress: 10%  [|_________]
			if(/^Progress: (.*?%)/) {
				$info{progress} = $1;

				if(/^Progress: .*?%\s+(\[.*?\])/) {
					$info{progress_bar} = $1;
				}
			}
		}
		close F;
	}
	
	return %info;
}

sub parse_clientcfg {
	my $file = $_[0];

	print "Parsing client.cfg file: $file\n" if($cfg{verbose});
	
	my %info;
	if(-r $file) {
		my $section = 'undefined';
	
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			# strip windows carriage return character
			if(s/\r//g) {}

			#[settings]
			#username=[DPC]_Fatal_Error_Group0smoking2000
			#team=92
			#asknet=no
			#machineid=1
			#local=217
			#bigpackets=yes
			#passkey=1a2b6789bc2342353653afd2c65u987b
			#extra_parms=-smp 2 -verbosity 9 -forceasm
			#
			#[http]
			#active=no
			#host=localhost
			#port=8080
			#usepasswd=yes
			#proxy_name=pu
			#proxy_passwd=Ԯ
			#usereg=yes
			#
			#[power]
			#battery=yes
			#
			#[core]
			#priority=96
			#disableassembly=yes
			#checkpoint=30
			#ignoredeadlines=yes
			#
			#[clienttype]
			#memory=512
			#type=1
			#nonet=yes

			#[settings]
			if(/^\[(.*?)\]$/) {
				$section = $1;
			}
			#username=[DPC]_Fatal_Error_Group0smoking2000
			if(/^(.*?)=(.*)$/) {
				$info{by_section}{$section}{lc($1)} = $2;
				$info{by_key}{lc($1)} = $2;
			}
		}
		close F;
	}

	return %info;
}

sub parse_fahlog {
	my $file = $_[0];

	print "Parsing FAHlog.txt file: $file\n" if($cfg{verbose});
	
	my %info;
	if(-r $file) {
		my $i    = 0;
		my $data = '';
		my @line = ();
	
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;

			# strip windows carriage return character
			if(s/\r//g) {}
			
			# save the line in the list
			push @line, $_;

			# add the line with line number to variable 
			# to use in regular expressions
			$data .= "$i:".$_."\n";

			# increment the line counter
			$i++;
		}
		
		$i = 0;
		my $current   = '';
		my $previous  = '';
		my $last_date = '';

                my @steps = ();
                my %seen  = ();

                my $time  = '';
                my $steps_completed = 0;
                my $steps_total     = 0;
                my $percentage      = 0;

		foreach(@line) {
			$previous = $current;
			$current  = $_;
			
			#########################
			# Parse log header
			#
			# Example:
			#
			#
			# --- Opening Log file [June 7 10:28:47]
			# 
			# 
			# # Linux Console Edition #######################################################
			# ###############################################################################
			# 
			#                        Folding@Home Client Version 5.02
			# 
			#                           http://folding.stanford.edu
			# 
			# ###############################################################################
			# ###############################################################################
			# 
			# Launch directory: /home/bas/software/folding-schapie
			# Executable: ./FAH502-Linux.exe
			# Arguments: -verbosity 9 -oneunit
			# 
			
			# --- Opening Log file [June 7 10:28:47]
			if(/^.*? Opening Log file \[(.*)\]/) {
				# whenever the client is (re)started it will print this header in the log
				# the client can be restarted many times in a single log 

				$info{file_opened} = $1;
			}
			
			# Arguments: -verbosity 9 -configonly
			if(/^Arguments:\s*(.*)\s*/) {
				my $s = $1;
				if($s =~ s/^(.*?)\s*$/$1/){}
				if($s =~ s/-(be|in)\w+//g){}

				#print "arguments : $s\n";
				$info{client}{arguments} = $s;
			}
			
			# Executable: FAH502-Linux.exe
			if(/^Executable:\s*(.*)\s*/) {
				#print "executable : $1\n";
				$info{client}{executable} = $1;
			}
			
			# Launch directory: /home/bas/software/folding
			if(/^Launch directory:\s*(.*)\s*/) {
				#print "directory  : $1\n";
				$info{client}{directory} = $1;
			}
			
			#                        Folding@Home Client Version 5.02
			#                        Folding@Home Client Version 5.90alpha                        
			if(/Folding\@Home Client Version (.*)\s*/) {
				#print "version    : $1\n";
				$info{client}{version} = $1;
			}
			
			# # Windows Console GPU Edition ######
			# # Linux Console Edition ######
			if(/^#\s+(.*)\s+###/) {
				#print "type       : $1\n";
				$info{client}{type} = $1;
			}

		
			#########################
			# Parse client startup & core initialization
			#
			# [19:09:29] - Ask before connecting: No
			# [19:09:29] - User name: [DPC]_Fatal_Error_Group0smoking2000 (Team 92)
			# [19:09:29] - User ID: 722F245B4DE04A25
			# [19:09:29] - Machine ID: 1
			# [19:09:29]
			# [19:09:30] Loaded queue successfully.
			# [19:09:30] + Benchmarking ...
			# [19:09:37] The benchmark result is 437
			# [19:09:37] - Autosending finished units...
			# [19:09:37] Trying to send all finished work units
			# [19:09:37] + No unsent completed units remaining.
			# [19:09:37] - Autosend completed
			# [19:09:37]
			# [19:09:37] + Processing work unit
			# [19:09:37] Core required: FahCore_a0.exe
			# [19:09:37] Core found.
			# [19:09:37] Working on Unit 07 [December 18 19:09:37]
			# [19:09:37] + Working ...
			# [19:09:37] - Calling './FahCore_a0.exe -dir work/ -suffix 07 -checkpoint 15 -verbose -lifeline 25259 -version 502'
			# 

			# [19:09:29] - Ask before connecting: No
			if(/^\[(\d+:\d+:\d+)\] - Ask before connecting: (.*?)$/) {
				$info{config}{ask_before_connecting}{time}  = $1;
				$info{config}{ask_before_connecting}{value} = $2;
			}

			# [16:37:02] - Use IE connection settings: Yes
			if(/^\[(\d+:\d+:\d+)\] - Use IE connection settings: (.*?)$/) {
				$info{config}{use_ie_connection_settings}{time}  = $1;
				$info{config}{use_ie_connection_settings}{value} = $2;
			}
			
			# [19:09:29] - User name: [DPC]_Fatal_Error_Group0smoking2000 (Team 92)
			if(/^\[(\d+:\d+:\d+)\] - User name: (.*?) \(Team (\d+)\)$/) {
				$info{config}{username}{time}  = $1;
				$info{config}{username}{value} = $2;
				
				$info{config}{team}{time}      = $1;
				$info{config}{team}{value}     = $3;
			}
			
			# [19:09:29] - User ID: 722F245B4DE04A25
			if(/^\[(\d+:\d+:\d+)\] - User ID: (.*?)$/) {
				$info{config}{userid}{time}  = $1;
				$info{config}{userid}{value} = $2;
			}
			# [14:15:01] - User ID not found locally
			# [14:15:01] + Requesting User ID from server
			#
			# [14:15:01] - User ID not found locally
			elsif(/^\[(\d+:\d+:\d+)\] - User ID not found locally$/) {
				$info{config}{userid_not_found_locally}{time} = $1;
			}
			# [14:15:01] + Requesting User ID from server
			elsif(/^\[(\d+:\d+:\d+)\] + Requesting User ID from server$/) {
				$info{config}{requesting_userid_from_server}{time} = $1;
			}
			# [10:54:20] - User ID not found locally
			# [10:54:20] + Requesting User ID from server
			# [10:54:20] - Getting ID from AS:
			# [10:54:20] Connecting to http://assign.stanford.edu:8080/
			# [10:54:21] Posted data.
			# [10:54:21] Initial: 7AC9; - Received User ID = 7AC9CAF93C9E3C9B
			#
			# [10:54:20] - Getting ID from AS: 
			elsif(/^\[(\d+:\d+:\d+)\] - Getting ID from AS: (.*)$/) {
				$info{config}{getting_id_from_as}{time}  = $1;
				$info{config}{getting_id_from_as}{value} = $2 ? $2 : "";
			}
			# [10:54:20] Connecting to http://assign.stanford.edu:8080/
			elsif(/^\[(\d+:\d+:\d+)\] Connecting to (.*?)$/) {
				$info{config}{connecting_to}{time}  = $1;
				$info{config}{connecting_to}{value} = $2;
			}
			# [10:54:21] Posted data.
			elsif(/^\[(\d+:\d+:\d+)\] Posted data.$/) {
				$info{config}{posted_data}{time} = $1;
			}
			# [10:54:21] Initial: 7AC9; - Received User ID = 7AC9CAF93C9E3C9B
			elsif(/^\[(\d+:\d+:\d+)\] Initial: (\w+); - Received User ID = (\w+)$/) {
				$info{config}{initial}{time}  = $1;
				$info{config}{initial}{value} = $2;
				
				$info{config}{received_userid}{time}  = $1;
				$info{config}{received_userid}{value} = $3;
			}
			# [14:12:08] - User ID not found locally
			# [14:12:08] + Requesting User ID from server
			# [14:12:36] + Could not connect to Primary Assignment Server for ID
			# [14:13:04] + Could not connect to Secondary Assignment Server for ID
			# [14:13:04] 
			# + Could not get ID from server. Retrying...
			
			# [14:12:36] + Could not connect to Primary Assignment Server for ID
			elsif(/^\[(\d+:\d+:\d+)\] \+ Could not connect to Primary Assignment Server for ID$/) {
				$info{config}{could_not_connect_to_primary_assignment_server_for_id}{time} = $1;
			}
			# [14:13:04] + Could not connect to Secondary Assignment Server for ID
			elsif(/^\[(\d+:\d+:\d+)\] \+ Could not connect to Secondary Assignment Server for ID$/) {
				$info{config}{could_not_connect_to_secondary_assignment_server_for_id}{time} = $1;
			}
			# [14:13:04] 
			elsif(/^\[(\d+:\d+:\d+)\] $/) {
				$info{config}{empty_line}{time} = $1;
			}
			# + Could not get ID from server. Retrying...
			elsif(/^\+ Could not get ID from server. Retrying\.\.\.$/) {
				if($previous =~ /^\[(\d+:\d+:\d+)\] $/) {
					$info{config}{could_not_get_id_from_server_retrying}{time} = $1;
				}
			}

			# [19:09:30] Loaded queue successfully.
			# [19:09:30] + Benchmarking ...
			# [19:09:37] The benchmark result is 437
			
			# [19:09:30] Loaded queue successfully.
			if(/^\[(\d+:\d+:\d+)\] Loaded queue successfully\.$/) {
				$info{client_initialization}{loaded_queue_successfully}{time} = $1;
			}
			# [19:09:30] + Benchmarking ...
			if(/^\[(\d+:\d+:\d+)\] \+ Benchmarking \.\.\.$/) {
				$info{client_initialization}{benchmarking}{time} = $1;
			}
			# [19:09:37] The benchmark result is 437
			if(/^\[(\d+:\d+:\d+)\] The benchmark result is (.*?)$/) {
				$info{client_initialization}{benchmark_result}{time}  = $1;
				$info{client_initialization}{benchmark_result}{value} = $2;
			}
			
			# [19:09:37] - Autosending finished units...
			# [19:09:37] Trying to send all finished work units
			# [19:09:37] + No unsent completed units remaining.
			# [19:09:37] - Autosend completed
			
			# [19:09:37] - Autosending finished units...
			if(/^\[(\d+:\d+:\d+)\] - Autosending finished units\.\.\.$/) {
				$info{client_initialization}{autosending_finished_units}{time} = $1;
			}
			# [19:09:37] Trying to send all finished work units
			if(/^\[(\d+:\d+:\d+)\] Trying to send all finished work units$/) {
				$info{client_initialization}{trying_to_send_all_finished_work_units}{time} = $1;
			}
			# [19:09:37] + No unsent completed units remaining.
			if(/^\[(\d+:\d+:\d+)\] \+ No unsent completed units remaining\.$/) {
				$info{client_initialization}{no_unsent_completed_units_remaining}{time} = $1;
			}
			# [19:09:37] - Autosend completed
			if(/^\[(\d+:\d+:\d+)\] - Autosend completed$/) {
				$info{client_initialization}{autosend_completed}{time} = $1;
			}

			# [19:09:37] + Processing work unit
			# [19:09:37] Core required: FahCore_a0.exe
			# [19:09:37] Core found.
			# [19:09:37] Working on Unit 07 [December 18 19:09:37]
			# [19:09:37] + Working ...
			# [19:09:37] - Calling './FahCore_a0.exe -dir work/ -suffix 07 -checkpoint 15 -verbose -lifeline 25259 -version 502'
			#
			# [18:20:33] + Processing work unit
			# [18:20:33] Core required: FahCore_a0.exe
			# [18:20:33] Core not found.
			# [18:20:33] - Core is not present or corrupted.
			# [18:20:33] - Attempting to download new core...
			 
			# [19:09:37] + Processing work unit
			if(/^\[(\d+:\d+:\d+)\] \+ Processing work unit$/) {
				$info{workunit_initialization}{processing_work_unit}{time} = $1;
			}
			
			# [19:09:37] Core required: FahCore_a0.exe
			if(/^\[(\d+:\d+:\d+)\] Core required: (.*?)$/) {
				$info{workunit_initialization}{core_required}{time}  = $1;
				$info{workunit_initialization}{core_required}{value} = $2;
			}
			
			# [19:09:37] Core found.
			if(/^\[(\d+:\d+:\d+)\] Core found\.$/) {
				$info{workunit_initialization}{core_found}{time} = $1;
			}
			
			# [19:09:37] Working on Unit 07 [December 18 19:09:37]
			if(/^\[(\d+:\d+:\d+)\] Working on Unit (\d+) \[(.*?)\]$/) {
				$info{workunit_initialization}{working_on_unit}{time}  = $1;
				$info{workunit_initialization}{working_on_unit}{value} = $2;
				
				$info{workunit_initialization}{working_on_unit_date}{time}  = $1;
				$info{workunit_initialization}{working_on_unit_date}{value} = $3;
			}
			
			# [19:09:37] + Working ...
			if(/^\[(\d+:\d+:\d+)\] \+ Working \.\.\.$/) {
				$info{workunit_initialization}{working}{time} = $1;
			}
			
			# [19:09:37] - Calling './FahCore_a0.exe -dir work/ -suffix 07 -checkpoint 15 -verbose -lifeline 25259 -version 502'
			if(/^\[(\d+:\d+:\d+)\] - Calling '(.*?)'$/) {
				$info{workunit_initialization}{calling}{time}  = $1;
				$info{workunit_initialization}{calling}{value} = $2;
			}
			
			# [18:20:33] Core not found.
			if(/^\[(\d+:\d+:\d+)\] Core not found\.$/) {
				$info{workunit_initialization}{core_not_found}{time} = $1;
			}
			# [18:20:33] - Core is not present or corrupted.
			elsif(/^\[(\d+:\d+:\d+)\] - Core is not present or corrupted\.$/) {
				$info{workunit_initialization}{core_not_present_or_corrupted}{time} = $1;
			}
			# [18:20:33] - Attempting to download new core...
			elsif(/^\[(\d+:\d+:\d+)\] - Attempting to download new core\.\.\.$/) {
				$info{workunit_initialization}{attempting_to_download_new_core}{time} = $1;
			}

			# [18:20:33] + Downloading new core: FahCore_a0.exe
			# [18:20:33] Downloading core (/~pande/Linux/x86//Core_a0.fah from www.stanford.edu)
			# [18:21:01] Could not download 'www.stanford.edu'
			# [18:21:01] + Error: Could not download core
			# [18:21:01] + Core download error (#2), waiting before retry...
			
			# [18:20:33] + Downloading new core: FahCore_a0.exe
			if(/^\[(\d+:\d+:\d+)\] \+ Downloading new core: (.*?)$/) {
				$info{download_new_core}{downloading_new_core}{time}  = $1;
				$info{download_new_core}{downloading_new_core}{value} = $2;
			}
			# [18:20:33] Downloading core (/~pande/Linux/x86//Core_a0.fah from www.stanford.edu)
			elsif(/^\[(\d+:\d+:\d+)\] Downloading core \((.*?)\)$/) {
				$info{download_new_core}{downloading_core}{time}  = $1;
				$info{download_new_core}{downloading_core}{value} = $2;
			}
			# [18:21:01] Could not download 'www.stanford.edu'
			elsif(/^\[(\d+:\d+:\d+)\] Could not download '(.*?)'$/) {
				$info{download_new_core}{could_not_download}{time}  = $1;
				$info{download_new_core}{could_not_download}{value} = $2;
			}
			# [18:21:01] + Error: Could not download core
			elsif(/^\[(\d+:\d+:\d+)\] \+ Error: Could not download core$/) {
				$info{download_new_core}{error_could_not_download_core}{time} = $1;
			}
			# [18:21:01] + Core download error (#2), waiting before retry...
			elsif(/^\[(\d+:\d+:\d+)\] \+ Core download error \((.*?)\), waiting before retry\.\.\.$/) {
				$info{download_new_core}{core_download_error_waiting_before_retry}{time}  = $1;
				$info{download_new_core}{core_download_error_waiting_before_retry}{value} = $2;
			}
			

			#########################
			# Parse core shutdown
			#
			# [06:04:25] Folding@home Core Shutdown: FINISHED_UNIT
			# [06:04:28] CoreStatus = 64 (100)
			#
			# [19:25:17] Folding@home Core Shutdown: EARLY_UNIT_END
			# [19:25:19] CoreStatus = 72 (114)
			#
			# [19:54:18] GPU was interrupted
			# [19:54:18] Folding@home Core Shutdown: INTERRUPTED
			# [19:54:26] CoreStatus = 66 (102)
			#
			# [12:00:24] Folding@home Core Shutdown: CORE_OUTDATED
			# [12:00:28] CoreStatus = 6E (110)
			#
			# [12:34:56] Folding@home Core Shutdown: BAD_WORK_UNIT
			# [12:35:00] CoreStatus = 72 (114)
			#
			# [12:34:56] Folding@home Core Shutdown: UNKNOWN_ERROR 
			# [12:35:00] CoreStatus = 61 (97) 
			#
			# [12:34:56] Folding@home Core Shutdown: BAD_FILE_FORMAT
			# [12:35:00] CoreStatus = 6F (111)
			#
			# [12:34:56] Folding@home Core Shutdown: BAD_FRAME_CHECKSUM
			# [12:35:00] CoreStatus = 70 (112)
			#
			# [12:34:56] Folding@home Core Shutdown: FILE_IO_ERROR
			# [12:35:00] CoreStatus = 75 (117)
			#
			# [12:34:56] Folding@home Core Shutdown: UNSTABLE_MACHINE
			# [12:35:00] CoreStatus = 7A (122)

			if(/^Folding\@home Core Shutdown: (\w+)/) {
				$info{core_shutdown}{time}  = $1;
				$info{core_shutdown}{value} = $2;
			}
			
			#########################
			# Parse client shutdown
			#
			# Folding@Home Client Shutdown.

			if(/^Folding\@Home Client Shutdown\./) {
				$info{client_shutdown}{line} = $i + 1;
			}
			
			$i++;
		}
		close F;

		if(exists $info{client_shutdown} && exists $info{client_shutdown}{line} && $info{client_shutdown}{line} != $i) {
			$info{client_shutdown}{continued} = 1;
		}
	}

	return %info;
}

sub parse_myfolding {
	my $file = $_[0];
	
	print "Parsing MyFolding.html file: $file\n" if($cfg{verbose});

	my %info;
	if(-r $file) {
		open(F, $file) || die "Cannot open file ($!)";
		while(<F>) {
			chomp;
			
			#User name: [DPC]_Fatal_Error_Group0smoking2000<br>
			if(/^User name: (.*?)\<br\>/) {
				$info{username} = $1;
			}
			
			#Team number: 92<br>
			if(/^ Team number: (\d+)\<br\>/) {
				$info{team} = $1;
			}
			
			#Personal statistics: <a href="http://vspx27.stanford.edu/cgi-bin/main.py?qtype=userpage&username=$USERNAME&teamnum=$TEAM">http://vspx27.stanford.edu/cgi-bin/main.py?qtype=userpage&username=[DPC]_Fatal_Error_Group0smoking2000&teamnum=92</a><br>
			if(/^Personal statistics: \<a href="(.*?)"\>.*?<\/a\>\<br\>/) {
				$info{statistics}->{userpage} = $1;
			}
			
			#Team statistics: <a href="http://vspx27.stanford.edu/cgi-bin/main.py?qtype=teampage&teamnum=92">http://vspx27.stanford.edu/cgi-bin/main.py?qtype=teampage&teamnum=92</a><br>
			if(/^Team statistics: \<a href="(.*?)"\>.*?<\/a\>\<br\>/) {
				$info{statistics}->{teampage} = $1;
			}
		}
		close F;
	}

	return %info;
}

sub parse_teamstats {
	my $file = shift;
	my $team = shift;

	if(!$team && $file && $file =~ /\/team(\d+)\.txt$/) {
		$team = $1;
	}

	print "Parsing teamstats file: $file\n" if($cfg{verbose});

	my %info = ();

	# parse the HTML
	if(-r $file) {
	        open (F, "<", $file) || die "Cannot open file: $file ($!)";
                while (<F>) {                           # read from file
                        # FOLDING@HOME TEAM STATS: Team # 92
                        #
                        #
                        # Wed Feb 15 12:00:00 PST 2006
                        # team  teamname        score   wu
                        # 92    Dutch Power Cows        38482957        437189
                        # Team rank: 19
                        #
                        # Team members:
                        # rank  team rank       name    credit  total   team
                        # 261   1       gowron  1346960 20515   92
                        # 445   2       Team_Pastis_51  935368  9995    92
                        # 489   3       DPC_Los_Alcoholicos0Popipipo    881083  6865    92

                        chomp;

                        #Wed Feb 15 12:00:00 PST 2006
                        if(/^(\w+\s+\w+\s+\d+\s+\d+:\d+:\d+\s+\w+\s+\d+)\s*$/) {
                                print "Generated     : $1\n" if($cfg{debug});
                                $info{generated} = $1;

                                #Wed Feb 15 12:00:00 PST 2006
                                if($info{generated} =~ /^\w+\s+(\w+)\s+(\d+)\s+(\d+:\d+:\d+)\s+(\w+)\s+(\d+)$/) {
                                        my $month = Date::Calc::Decode_Month($1);
                                        my $day   = $2;
                                        my $time  = $3;
                                        my $tzone = $4;
                                        my $year  = $5;

                                        my ($t_year, $t_month, $t_day) = Date::Calc::Today();

                                        $month   = '0'.$month   if(length($month)   == 1);
                                        $day     = '0'.$day     if(length($day)     == 1);
                                        $t_month = '0'.$t_month if(length($t_month) == 1);
                                        $t_day   = '0'.$t_day   if(length($t_day)   == 1);

                                        my $t_date = $t_year.$t_month.$t_day;
                                        my $date   = $year.$month.$day;
                                        my $diff   = $date - $t_date;

                                        # if there is more than 1 day difference in the date
                                        # the teamstats file was generated and todays date,
                                        # most likely the teams rank dropped to low, and no
                                        # new teamstats file is being generated anymore.
                                        # Or the date on the system is wrong.
                                        if($diff > 1 || $diff < -1) {
                                                print "The teamstats file is too old ($diff days).\n" if($cfg{debug});
                                        }
                                }
                        }
                        #92    Dutch Power Cows        38482957        437189
                        elsif(/^($team)\t(.*?)\t(\d+)\t(\d+)$/) {
                                if($cfg{debug}) {
                                        print "Team number   : $1\n";
					print "Team name     : $2\n";
					print "Team score    : $3\n";
					print "Team wu       : $4\n";
                                }

                                $info{team}->{number} = $1;
                                $info{team}->{name}   = $2;
                                $info{team}->{score}  = $3;
                                $info{team}->{wu}     = $4;
                        }
                        #Team rank: 19
                        elsif(/^Team rank: (\d+)$/) {
                                print "Team rank     : $1\n" if($cfg{debug});
                                $info{team}->{rank} = $1;
                        }
                        #Team members:
                        #if(/^(Team members:)$/) {
                        #       print $1."\n";
                        #}
                        #261   1       gowron  1346960 20515   92
                        elsif(/^(\d+)\t(\d+)\t(.*?)\t(\d+)\t(\d+)\t($team)$/) {
                                if($cfg{debug}) {
					print "\n";
                                        print "User rank     : $1\n";
					print "User team rank: $2\n";
					print "User name     : $3\n";
					print "User credit   : $4\n";
					print "User total    : $5\n";
					print "User team     : $6\n";
                                }

                                $info{rank_entry}->{$2}->{rank}      = $1;
                                $info{rank_entry}->{$2}->{team_rank} = $2;
                                $info{rank_entry}->{$2}->{name}      = $3;
                                $info{rank_entry}->{$2}->{credit}    = $4;
                                $info{rank_entry}->{$2}->{total}     = $5;
                                $info{rank_entry}->{$2}->{team}      = $6;
                        }
		}
		close F;

		$info{last_update} = get_date();
	}

	return %info;
}

sub parse_eoc_user_summary {
	my $file = shift;
	my %cfg  = @_;

	print " Parsing EOC User Summary file: $file\n" if($cfg{verbose});

	my %info;
	if(-r $file) {
		my $xml = XMLin( $file ) || die $@;

		%info = %{$xml};
	}
	
	return %info;
}

sub update_stanford_urls {
	my %username = @_;

	if(!$username{name} || !(exists $username{team} && $username{team} >= 0)) {
		print "Error: Cannot update Stanford URLs without the username and team!\n";
	}

	# userpage & teampage are only recorded in the MyFolding.html,
	# but the webserver url is hardcoded in the client and outdated on most v5 clients.
	# Overwrite the URLs to point to the correct webserver hostname.
	 
	my $userpage  = "http://fah-web.stanford.edu/cgi-bin/main.py?qtype=userpage&username=";
	   $userpage .= $username{name};
	   $userpage .= "&teamnum=";
	   $userpage .= $username{team};
	
	my $teampage  = "http://fah-web.stanford.edu/cgi-bin/main.py?qtype=teampage&teamnum=";
	   $teampage .= $username{team};
	
	$username{external_resources}->{userpage} = $userpage;
	$username{external_resources}->{teampage} = $teampage;
	
	$team_list->{team_list}->{team}->{$username{team}}->{external_resources}->{teampage} 
		= $teampage;
		
		
	# add username element to username_list
	my $userid = $username{team}.'#'.$username{name};
	$username{userid} = $userid;
		
	$username_list->{username_list}->{username}->{$userid} = \%username;
}

sub iso_date {
        my $date     = $_[0];
        my $iso_date = $date;

        # Tue Jan 17 13:37:47 2006
        if($date =~ /^(\w+)\s+(\w+)\s+(\d+)\s+(\d+:\d+:\d+)\s+(\d+)$/) {
                my $dow   = $1;
                my $month = $2;
                my $day   = $3;
                my $time  = $4;
                my $year  = $5;

                $month = Date::Calc::Decode_Month($month);

                $day   = '0'.$day   if(length($day)   == 1);
                $month = '0'.$month if(length($month) == 1);

                $iso_date = "$year-$month-$day $time";
        }

        return $iso_date;
}

sub get_tpf {
        my $project     = shift;
        my $run         = shift;
        my $clone       = shift;
        my $generation  = shift;
        my $begin       = shift;
	my $index       = shift;
	my $last_update = shift;
	my $lines       = shift;
	my %cfg         = @_;

	# Content of FAHlog-Complete
	my @lines = @{$lines};
	
	if($cfg{debug}) {
		print "Looking up Time Per Frame lines for:\n";
		print " Project: $project (Run $run, Clone $clone, Gen $generation)\n";
		print "Looking up index $index\n";
	}

	# extract frame times from FAHlog for this WU
	my $i = 0;
        my @steps = ();
        my %seen  = ();

	my $year = Date::Calc::This_Year();
	if($last_update =~ /^(\d{4})-\d{2}-\d{2} \d{2}:\d{2}:\d{2}$/) {
		$year = $1;
	}

	my $previous = '';
	my $current  = '';
	
        my $previous_time = '';
        my $current_time  = '';

	my $min = '';
	my $max = 0;
	my $tot = 0;
	my $avg = 0;
	my $cur = 0;
	
	my $line = 0;
        my $date = '';
        my $time = '';

        my $steps_completed = 0;
        my $steps_total     = 0;
        my $percentage      = 0;

	my $found_index   = 0;
        my $found_project = 0;
	my $prcg          = '';

	my $time_start = new Benchmark;
	foreach(@lines) {
		$line++;
		
		chomp;

		s/\r//g; # strip carriage return

		$previous = $current;
		$current  = $_;

		print "$_\n" if($cfg{debug});

		# String to search for:
		#
		# [18:27:34] Working on Unit 03 [June 4 18:27:34]
		# [01:19:48] Working on queue slot 09 [April 9 01:19:48 UTC]
		if(!$found_index) {
			# [06:50:50] Working on queue slot 01 [April 11 06:50:50 UTC]
			# [06:18:35] Working on Unit 04 [January 7 06:18:35]
			if(/^\[(\d{2}:\d{2}:\d{2})\] Working on (queue slot|Unit) $index \[(\w+)\s+(\d+)\s+(\d{2}):(\d{2}):(\d{2})( UTC)?\]\s*$/) {
				my $time  = $1;
				my $slot  = $2;
				my $month = $3;
				my $day   = $4;
				my $hour  = $5;
				my $min   = $6;
				my $sec   = $7;
				my $tzone = $8;

		                if($cfg{debug}) {
		                	print "Debug: Possibly found current index $index ($month $day $hour:$min:$sec)\n";
				}

				# Update the date to keep track on which date the FAHlog messages are logged.
				$month = 'February' if($month eq 'Feburary');
				$month = Date::Calc::Decode_Month($month);
				$month = '0'.$month if(length($month) == 1);
				$day   = '0'.$day   if(length($day)   == 1);
				$hour  = '0'.$hour  if(length($hour)  == 1);
				$min   = '0'.$min   if(length($min)   == 1);
				$sec   = '0'.$sec   if(length($sec)   == 1);

				print "Updating FAHlog date to: $year-$month-$day, was: $date\n" if($cfg{debug});
				$date = "$year-$month-$day";

				print "Checking if date is within 24 hours from begin timestamp: $begin\n" if($cfg{debug});
				# Sun May  3 11:01:56 2009
				if($begin =~ /^(\w+)\s+(\w+)\s+(\d+)\s+(\d{2}):(\d{2}):(\d{2})\s+(\d{4})$/) {
					my $begin_dow   = $1;
					my $begin_month = $2;
					my $begin_day   = $3;
					my $begin_hour  = $4;
					my $begin_min   = $5;
					my $begin_sec   = $6;
					my $begin_year  = $7;
					
					$begin_month = Date::Calc::Decode_Month($begin_month);

					my ($min_year, $min_month, $min_day, $min_hour, $min_min, $min_sec)
						= Date::Calc::Add_Delta_DHMS( $begin_year, $begin_month, $begin_day, $begin_hour, $begin_min, $begin_sec,
									      0, -24, 0, 0
									    );
					$min_month = '0'.$min_month if(length($min_month) == 1);
					$min_day   = '0'.$min_day   if(length($min_day)   == 1);
					$min_hour  = '0'.$min_hour  if(length($min_hour)  == 1);
					$min_min   = '0'.$min_min   if(length($min_min)   == 1);
					$min_sec   = '0'.$min_sec   if(length($min_sec)   == 1);

					my ($max_year, $max_month, $max_day, $max_hour, $max_min, $max_sec)
						= Date::Calc::Add_Delta_DHMS( $begin_year, $begin_month, $begin_day, $begin_hour, $begin_min, $begin_sec,
									      0, 24, 0, 0
									    );
					$max_month = '0'.$max_month if(length($max_month) == 1);
					$max_day   = '0'.$max_day   if(length($max_day)   == 1);
					$max_hour  = '0'.$max_hour  if(length($max_hour)  == 1);
					$max_min   = '0'.$max_min   if(length($max_min)   == 1);
					$max_sec   = '0'.$max_sec   if(length($max_sec)   == 1);

					my $min_date = "$min_year-$min_month-$min_day $min_hour:$min_min:$min_sec";
					my $max_date = "$max_year-$max_month-$max_day $max_hour:$max_min:$max_sec";
	
					my $found_date = "$year-$month-$day $hour:$min:$sec";

					my $date1 = ParseDate($found_date);
					my $date2 = ParseDate($min_date);
					my $date3 = ParseDate($max_date);
					
					my $min_flag = Date_Cmp($date1, $date2);
					my $max_flag = Date_Cmp($date1, $date3);
					
					if($cfg{debug}) {
						print "Found: $found_date ($date1)\n";
						print "  Min: $min_date ($date2), Flag: $min_flag\n";
						print "  Max: $max_date ($date3), Flag: $max_flag\n";
					}

					# flag <  0: date1 is earlier than date2
					# flag == 0: date1 and date2 are identical
					# flag >  0: date1 is later than date2
					if($min_flag > 0 && $max_flag < 0) {
						# The found date is later than 24 hours before the begin time of the WU we're looking for, and
						# the found date is less than 24 hours after the begin time of the WU we're looking for.

						$found_index = 1;

	                                        if($cfg{debug}) {
	                                                print "Debug: Found current index $index ($month $day $hour:$min:$sec)\n";
	                                        }
					}
				}
				else {
					print "Warning: Cannot parse begin timestamp: $begin\n" if($cfg{debug});
				}
			}
		}
		else {
			# Try to keep track on which date the FAHlog messages are logged.
			#
			# Several logmesages contain the date which can be used to determin the date:
			#
			# --- Opening Log file [April 11 06:50:26 UTC]
			# --- Opening Log file [January 7 06:18:08]
			#
			# [06:50:27] - Autosending finished units... [April 11 06:50:27 UTC]
			# [08:28:22] - Autosending finished units... [08:28:22]
			#
			# [06:50:50] Working on queue slot 01 [April 11 06:50:50 UTC]
			# [06:18:35] Working on Unit 04 [January 7 06:18:35]
			#
			# [08:21:06] + Attempting to send results [April 12 08:21:06 UTC]
			
			# --- Opening Log file [April 11 06:50:26 UTC]
			# --- Opening Log file [January 7 06:18:08]
			#
			if(/^\-\-\- Opening Log file \[(\w+) (\d+) (\d{2}:\d{2}:\d{2})( UTC)?\]/) {
				my $month = $1;
				my $day   = $2;
				my $time  = $3;
	
				$month = 'February' if($month eq 'Feburary');
				$month = Date::Calc::Decode_Month($month);
				$month = '0'.$month if(length($month) == 1);
				$day   = '0'.$day   if(length($day)   == 1);
		
				print "Updating FAHlog date to: $year-$month-$day, was: $date\n" if($cfg{debug});
				$date = "$year-$month-$day";
			}
			# [06:50:27] - Autosending finished units... [April 11 06:50:27 UTC]
			# [06:50:27] - Autosending finished units... [April 11 06:50:27]
			elsif(/^\[(\d{2}:\d{2}:\d{2})\] \- Autosending finished units\.\.\. \[(\w+) (\d+) (\d{2}:\d{2}:\d{2})( UTC)?\]$/) {
				my $time  = $1;
				my $month = $2;
				my $day   = $3;
				my $time2 = $4;
	
				$month = 'February' if($month eq 'Feburary');
				$month = Date::Calc::Decode_Month($month);
				$month = '0'.$month if(length($month) == 1);
				$day   = '0'.$day   if(length($day)   == 1);

				print "Updating FAHlog date to: $year-$month-$day, was: $date\n" if($cfg{debug});
				$date = "$year-$month-$day";
			}
			# [06:50:50] Working on queue slot 01 [April 11 06:50:50 UTC]
			# [06:18:35] Working on Unit 04 [January 7 06:18:35]
			elsif(/^\[(\d{2}:\d{2}:\d{2})\] Working on (queue slot|Unit) \d{2} \[(\w+) (\d+) (\d{2}:\d{2}:\d{2})( UTC)?\]$/) {
				my $time  = $1;
				my $slot  = $2;
				my $month = $3;
				my $day   = $4;
				my $time2 = $5;
	
				$month = 'February' if($month eq 'Feburary');
				$month = Date::Calc::Decode_Month($month);
				$month = '0'.$month if(length($month) == 1);
				$day   = '0'.$day   if(length($day)   == 1);

				print "Updating FAHlog date to: $year-$month-$day, was: $date\n" if($cfg{debug});
				$date = "$year-$month-$day";
			}
			# [08:21:06] + Attempting to send results [April 12 08:21:06 UTC]
			elsif(/^\[(\d{2}:\d{2}:\d{2})\] \+ Attempting to send results \[(\w+) (\d+) (\d{2}:\d{2}:\d{2}) UTC\]$/) {
				my $time  = $1;
				my $month = $2;
				my $day   = $3;
				my $time2 = $4;
	
				$month = 'February' if($month eq 'Feburary');
				$month = Date::Calc::Decode_Month($month);
				$month = '0'.$month if(length($month) == 1);
				$day   = '0'.$day   if(length($day)   == 1);

				print "Updating FAHlog date to: $year-$month-$day, was: $date\n" if($cfg{debug});
				$date = "$year-$month-$day";
			}
			# Detect day changes, if we cannot determin the date from the log messages
			elsif(/^\[(\d{2}:\d{2}:\d{2})\] /) {
				my $time = $1;
	
				$previous_time = $current_time;
				$current_time  = $time;
				
				my $date1 = ParseDate($date.' '.$previous_time);
				my $date2 = ParseDate($date.' '.$current_time);
				my $flag  = Date_Cmp($date1, $date2);
				
				if($flag < 0) {
					# date1 is earlier, the previous is older than the current
					#
					# the situation may have been:
	                                #  date1:     today 00:01:02 (previous: 00:01:02)
	                                #  date2:     today 00:02:03 (current : 00:02:03)
	        		}
				elsif($flag == 0) {
					# the two dates are identical
				}
				else {
					# date2 is earlier, the current is older than the previous
	                                # the situation may have been:
	                                #  date1: yesterday 23:59:58 (previous: 23:59:58)
	                                #  date2:     today 00:01:02 (current : 00:01:02)
	
					my ($year, $month, $day) = split '-', $date;
					   ($year, $month, $day) = Date::Calc::Add_Delta_Days($year, $month, $day, 1);
	
					$month = '0'.$month if(length($month) == 1);
					$day   = '0'.$day   if(length($day)   == 1);
	
					if($cfg{debug}) {
	                                	print "Warning: The current time is earlier than the previous ($current_time < $previous_time)\n";
						print "The previous date was probably yesterday at $previous_time.\n";
					}
				
					print "Updating FAHlog date to: $year-$month-$day, was: $date\n" if($cfg{debug});
					$date = "$year-$month-$day";
				}
			}
	
                        # project string to search for:
                        #
                        # [13:30:42] Project: 1496 (Run 24, Clone 10, Gen 4)
                        #
                        # or for the TINKER core:
                        #
                        # [20:51:00] Protein: p1152_L939_K12M_ext_from638
                        # [20:51:00] - Run: 76 (Clone 46, Gen 2)
                        # [20:51:00] - Frames Completed: 46, Remaining: 354
                        # [20:51:00] - Dynamic steps required: 4425000
                        if(!$found_project) {
                                if(/^\[(\d{2}:\d{2}:\d{2})\] Project: $project \(Run $run, Clone $clone, Gen $generation\)$/) {
                                        $found_project = 1;
                                        $prcg = "$project ($run, $clone, $generation)";

                                        if($cfg{debug}) {
                                                print "Debug: Found project $prcg\n";
                                        }
                                }
                                elsif($prcg eq $project && /^\[(\d{2}:\d{2}:\d{2})\] - Run: $run \(Clone $clone, Gen $generation\)$/) {
                                        $found_project = 1;
                                        $prcg .=  "$project ($run, $clone, $generation)";

                                        if($cfg{debug}) {
                                                print "Debug: Found project $prcg\n";
                                        }
                                }
                        }
			else {
                                # Gromacs based cores and Amber/PMD core:
                                #
                                # [09:16:00] Completed 311670 out of 500000 steps  (62)
                                # [10:36:55] Completed 315000 out of 500000 steps  (63)
                                #
                                # [09:16:00] Completed 311670 out of 500000 steps  (62%)
                                # [10:36:55] Completed 315000 out of 500000 steps  (63%)
                                #
                                # [09:16:00] Completed 311670 out of 500000 steps  (62 percent)
                                # [10:36:55] Completed 315000 out of 500000 steps  (63 percent)
				#
				# [03:14:06] Completed 42000 out of 200000 steps (21%)
                                #
                                if(/^\[(\d{2}:\d{2}:\d{2})\] Completed (\d+) out of (\d+) steps\s+\((\d+).*?\)$/) {
                                        $time            = $1;
                                        $steps_completed = $2;
                                        $steps_total     = $3;
                                        $percentage      = $4;

                                        if($cfg{debug}) {
                                                print "Debug: Date: $date, Time: $time; Steps completed: $steps_completed, total: $steps_total; Percentage: $percentage\n";
                                        }
                                        # if this is a previously seen step, skip it
                                        if($seen{$steps_completed}) {
                                                print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{debug});
                                                next;
                                        }
	                                
					my $current_time  = UnixDate($date.' '.$time.' UTC', "%s");
			                my $current_date  = ParseDateString("epoch $current_time");

					# skip the first time as we can't compare it to his predecessor
	                                my $seconds = -1;
	                                if($#steps >= 0) {
						my $previous_time = UnixDate($steps[$#steps]->{date} .' '. $steps[$#steps]->{time} .' UTC', "%s");
			                        my $previous_date = ParseDateString("epoch $previous_time");
						
	                                	$seconds = $current_time - $previous_time;
	
	                                	print "Difference between $previous_date and $current_date is $seconds.\n" if($cfg{debug});
	                                	
						if($seconds > $max) {
							$max = $seconds;
						}
						if($min eq '' || ($seconds < $min)) {
							$min = $seconds;
						}
						
						$tot += $seconds;
						$cur  = $seconds;
	                                }

                                        push @steps, { time       => $time,
                                                       completed  => $steps_completed,
                                                       total      => $steps_total,
                                                       percentage => $percentage,
						       date       => $date,
						       line       => $line,
						       seconds    => $seconds,
						       epoch      => $current_time,
                                                     };

                                        # flag this step as seen
                                        $seen{$steps_completed} = $time;

                                        $i++;
                                }
                                # GPU core:
                                #
                                # [05:39:50] Completed 10
                                #
                                # GPU work/logfile_??.txt can contain:
                                #
                                # Completed 10%
                                #
                                elsif(/^\[(\d{2}:\d{2}:\d{2})\] Completed (\d+)(\%|)$/) {
                                        my $s1 = $1;
                                        my $s2 = $2;
                                        my $s3 = $3;

                                        $time            = $s1;
                                        $steps_completed = $s2;
                                        $steps_total     = 100;
                                        $percentage      = $s2;

                                        if($cfg{debug}) {
                                                print "Debug: Date: $date, Time: $time; Steps completed: $steps_completed, total: $steps_total; Percentage: $percentage\n";
                                        }
                                        # if this is a previously seen step, skip it
                                        if($seen{$steps_completed}) {
                                                print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{debug});
                                                next;
                                        }

					my $current_time  = UnixDate($date.' '.$time.' UTC', "%s");
			                my $current_date  = ParseDateString("epoch $current_time");

					# skip the first time as we can't compare it to his predecessor
	                                my $seconds = -1;
	                                if($#steps >= 0) {
						my $previous_time = UnixDate($steps[$#steps]->{date}.' '.$steps[$#steps]->{time}.' UTC', "%s");
			                        my $previous_date = ParseDateString("epoch $previous_time");
						
	                                	$seconds = $current_time - $previous_time;
	
	                                	print "Difference between $previous_date and $current_date is $seconds. ($i)\n" if($cfg{debug});
	                                	
						if($seconds > $max) {
							$max = $seconds;
						}
						if($min eq '' || ($seconds < $min)) {
							$min = $seconds;
						}
						
						$tot += $seconds;
						$cur  = $seconds;
	                                }

                                        push @steps, { time       => $time,
                                                       completed  => $steps_completed,
                                                       total      => $steps_total,
                                                       percentage => $percentage,
						       date       => $date,
						       line       => $line,
						       seconds    => $seconds,
						       epoch      => $current_time,
                                                     };

                                        # flag this step as seen
                                        $seen{$steps_completed} = $time;

                                        $i++;
                                }
                                # TINKER core:
                                #
                                # [12:37:33] Protein: p1152_L939_K12M_ext_from638
                                # [12:37:33] - Run: 43 (Clone 14, Gen 22)
                                # [12:37:33] - Frames Completed: 5, Remaining: 395
                                # [12:37:33] - Dynamic steps required: 4937500
                                # ...
                                # [17:28:23] Finished a frame (12)
                                # [18:10:00] Finished a frame (13)
                                #
                                elsif(/Frames Completed: (\d+), Remaining: (\d+)$/ ||
                                      /^\[(\d{2}:\d{2}:\d{2})\] Finished a frame \((\d+)\)$/
                                ) {
                                        my $s1 = $1;
                                        my $s2 = $2;

                                        print "Debug: Date: $date  S1: $s1  S2: $s2\n" if($cfg{debug});

                                        # The first regex match is on a sigle digit,
                                        # so we're parsing a "Frames Completed: ..." line
                                        if($s1 =~ /^\d+$/) {
                                                $steps_completed = $s1;
                                                $steps_total     = $s1 + $s2;
                                        }
                                        # the first regex match is a timestamp
                                        # so we're parsing a "[00:00:00] Finished a frame (0)" line
                                        elsif($s1 =~ /^\d{2}:\d{2}:\d{2}$/) {
                                                $time            = $s1;
                                                $steps_completed = $s2;

                                                # if this is a previously seen step, skip it
                                                if($seen{$steps_completed}) {
                                                        print "Already seen step: $steps_completed (". ($i-1) ."), skipping.\n" if($cfg{debug});
                                                        next;
                                                }

						my $current_time = UnixDate($date.' '.$time.' UTC', "%s");
				                my $current_date = ParseDateString("epoch $current_time");

						# skip the first time as we can't compare it to his predecessor
		                                my $seconds = -1;
		                                if($#steps >= 0) {
							my $previous_time = UnixDate($steps[$#steps]->{date}.' '.$steps[$#steps]->{time}.' UTC', "%s");
				                        my $previous_date = ParseDateString("epoch $previous_time");
							
		                                	$seconds = $current_time - $previous_time;
		
		                                	print "Difference between $previous_date and $current_date is $seconds. ($i)\n" if($cfg{debug});
		                                	
							if($seconds > $max) {
								$max = $seconds;
							}
							if($min eq '' || ($seconds < $min)) {
								$min = $seconds;
							}
							
							$tot += $seconds;
							$cur  = $seconds;
		                                }
	
	                                        push @steps, { time       => $time,
	                                                       completed  => $steps_completed,
	                                                       total      => $steps_total,
	                                                       percentage => $percentage,
							       date       => $date,
							       line       => $line,
							       seconds    => $seconds,
						       	       epoch      => $current_time,
	                                                     };

                                                # flag this step as seen
                                                $seen{$steps_completed} = $time;
	
                                                $i++;

                                                last if($steps_completed == $steps_total);
                                        }
				}	
			}
		}
	}
	my $time_stop = new Benchmark;
	my $time_diff = timediff($time_stop, $time_start);

	print "File read in: ". timestr($time_diff) ."\n" if($cfg{debug});

	print "Found $i lines with progress information.\n" if($cfg{debug});

	if($i > 0) {
		# only compute the average TPF if we have at least 1 value
		$avg = sprintf("%.0f", $tot / ($i-1)) if($i > 1);

		if($cfg{debug}) {
			print "Max: $max ";
			print time_format($max);
			print "\n";
			print "Min: $min ";
			print time_format($min);
			print "\n";
			print "Tot: $tot ";
			print time_format($tot);
			print "\n";
			print "Avg: $avg ";
			print time_format($avg);
			print "\n";
			print "Cur: $cur ";
			print time_format($cur);
			print "\n";
		}

		return($min, $max, $avg, $cur, $tot, $i, @steps);
	}
	else {
		print "Warning: No useful log messages found!\n" if($cfg{debug});
	}

	return($min, $max, $avg, $cur, $tot, $i, @steps);
}

sub generate_flot_data {
	my $data = shift;
	my %cfg  = @_;

	my @steps = @{$data};

	my $flot_data = '';
	my $flot_data_time = '';
	my $flot_data_line = '';

	my $i = 0;
		
	my $time_start = new Benchmark;
	foreach my $step (@steps) {
		if($cfg{debug}) {
			print "Step $i:\n";
			print Dumper($step);
		}

		if($i == 0) {
			print "Skipping first step, which has no TPF data.\n" if($cfg{debug});
			$i++;
			next;
		}
		
		# The first flot data point is frame 2 (2%), the time between frame 1 (1%) and frame 2 (2%)
		$flot_data .= '['. ($i+1) .','. $step->{seconds} .']';
		$flot_data .= ','	  if($i < $#steps);
	
		$flot_data_time .= '0,0,' if($flot_data_time eq '');	
		$flot_data_time .= $step->{epoch} * 1000;
		$flot_data_time .= ','	  if($i < $#steps);

		$flot_data_line .= '0,0,' if($flot_data_line eq '');	
		$flot_data_line .= $step->{line};
		$flot_data_line .= ','	  if($i < $#steps);

	        $i++;
	}
	my $time_stop = new Benchmark;
	my $time_diff = timediff($time_stop, $time_start);
		
	print "Steps processed in: ". timestr($time_diff) ."\n" if($cfg{debug});

	return($flot_data, $flot_data_time, $flot_data_line);
}

sub get_expected {
	my $data = shift;
	my %cfg  = @_;

	my @steps = @{$data};

	# calculate the estimated time of completion of the current WU,
	# based on the average TPF of the last 3 frames
	
	my $steps_back  = 3;
	my $start_frame = ($#steps + 1) - $steps_back;

	if($start_frame >= 0) {
		my $recent_tot = 0;
		my $recent_avg = 0;

		my $i = $start_frame;

		while($i <= $#steps) {
			$recent_tot += $steps[$i]->{seconds};
			$i++;
		}

		$recent_avg = $recent_tot / $steps_back;

		my $steps_remaining = $steps[-1]->{total} - $steps[-1]->{completed};

		#my $time_remaining = $steps_remaining * $recent_avg;
		my $time_remaining = $steps_remaining * $cfg{steps_avg};

		my $time_expected = $steps[-1]->{epoch} + $time_remaining;
		my $date_expected = ParseDateString("epoch $time_expected");

		return $date_expected;
	}

	return 0;
}

sub previous_index {
	my $index = shift;

	$index = 9 if($index >= 10 || $index < 0);
	$index--;
	$index = 9 if($index == -1);

	return $index;
}

sub print_usage {
        my %cfg = @_;

        print "$cfg{self} v$cfg{version}\n\n";
        print "Usage: $cfg{self} [OPTIONS]\n\n";
        print "Options:\n";
        print "-d, --dir <DIR>       Path of the FCI server directory\n";
        print "-v, --verbose         Enable verbose output\n";
        print "-q, --quiet           Don't print any output at all, except for errors\n";
        print "-h, --help            Show this usage information\n";
}
